#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import copy
from itertools import product
from typing import Any
from unittest import mock
from unittest.mock import patch, PropertyMock

import numpy as np

import pandas as pd
import torch
from ax.adapter.cross_validation import AssessModelFitResult
from ax.adapter.registry import Generators
from ax.adapter.torch import TorchAdapter
from ax.core.arm import Arm
from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun
from ax.core.metric import Metric
from ax.core.objective import MultiObjective, Objective, ScalarizedObjective
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
)
from ax.core.outcome_constraint import ObjectiveThreshold, OutcomeConstraint
from ax.core.types import ComparisonOp
from ax.exceptions.core import DataRequiredError
from ax.generation_strategy.dispatch_utils import choose_generation_strategy_legacy
from ax.service.ax_client import AxClient
from ax.service.utils.best_point import (
    _extract_best_arm_from_gr,
    _is_row_feasible,
    derelativize_opt_config,
    get_best_by_raw_objective_with_trial_index,
    get_best_parameters_from_model_predictions_with_trial_index,
    get_best_raw_objective_point_with_trial_index,
    get_hypervolume_trace_of_outcomes_multi_objective,
    get_trace_by_arm_pull_from_data,
    get_values_of_outcomes_single_or_scalarized_objective,
    logger as best_point_logger,
)
from ax.service.utils.best_point_utils import select_baseline_name_default_first_trial
from ax.service.utils.instantiation import ObjectiveProperties
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_experiment,
    get_branin_metric,
    get_branin_search_space,
    get_experiment_with_map_data,
    get_experiment_with_observations,
)
from ax.utils.testing.mock import mock_botorch_optimize
from pyre_extensions import assert_is_instance, none_throws

best_point_module: str = get_best_by_raw_objective_with_trial_index.__module__
DUMMY_OPTIMIZATION_CONFIG = "test_optimization_config"


class TestBestPointUtils(TestCase):
    """Testing the best point utilities functionality that is not tested in
    main `AxClient` testing suite (`TestServiceAPI`)."""

    def test_get_values_of_outcomes_single_or_scalarized_objective(self) -> None:
        m1_value = 37
        m2_value = 49
        df = pd.DataFrame.from_records(
            [
                {"m1": m1_value, "m2": m2_value, "irrelevant": 4, "feasible": True},
                {"m1": 2, "m2": 3, "irrelevant": 4, "feasible": False},
            ]
        )
        metrics = [Metric(name="m1"), Metric(name="m2")]
        with self.subTest("MultiObjective not supported"):
            objective = MultiObjective(
                objectives=[Objective(metric=m, minimize=False) for m in metrics],
            )
            with self.assertRaisesRegex(ValueError, "MultiObjective is not supported"):
                get_values_of_outcomes_single_or_scalarized_objective(
                    df_wide=df, objective=objective
                )

        for minimize in [False, True]:
            objective = Objective(
                metric=Metric(name="m1", lower_is_better=minimize), minimize=minimize
            )
            worst = float("inf") if minimize else float("-inf")
            with self.subTest(f"Single objective, minimize={minimize}"):
                value = get_values_of_outcomes_single_or_scalarized_objective(
                    df_wide=df, objective=objective
                )
                self.assertIsInstance(value, np.ndarray)
                self.assertEqual(value.tolist(), [m1_value, worst])

            objective = ScalarizedObjective(
                metrics=metrics, weights=[0.4, 0.6], minimize=minimize
            )

            with self.subTest(f"ScalarizedObjective, {minimize=}"):
                value = get_values_of_outcomes_single_or_scalarized_objective(
                    df_wide=df, objective=objective
                )
                self.assertEqual(len(value), 2)
                self.assertAlmostEqual(value[0], 0.4 * m1_value + 0.6 * m2_value)
                self.assertEqual(value[1], worst)

            objective = ScalarizedObjective(
                metrics=metrics, weights=[0.4, -0.6], minimize=minimize
            )
            with self.subTest(
                f"ScalarizedObjective, {minimize=}, maximize other metric"
            ):
                value = get_values_of_outcomes_single_or_scalarized_objective(
                    df_wide=df, objective=objective
                )
                self.assertEqual(len(value), 2)
                self.assertEqual(value[0], 0.4 * m1_value - 0.6 * m2_value)
                self.assertEqual(value[1], worst)

    def test_get_hypervolume_trace_of_outcomes_multi_objective(self) -> None:
        objective = MultiObjective(
            objectives=[
                Objective(metric=Metric("m1"), minimize=False),
                Objective(metric=Metric("m2"), minimize=False),
            ],
        )
        # Objective thresholds will be inferred to be [1, 1]
        df_wide = pd.DataFrame.from_records(
            [
                {"m1": 1.0, "m2": 1.0, "feasible": True},
                {"m1": 2.0, "m2": 3.0, "feasible": True},
                {"m1": 4.0, "m2": 4.0, "feasible": False},
                {"m1": 3.0, "m2": 2.0, "feasible": True},
            ]
        )
        with self.subTest("Relative objective thresholds not supported"):
            optimization_config = MultiObjectiveOptimizationConfig(
                objective=objective,
                objective_thresholds=[
                    ObjectiveThreshold(
                        metric=Metric("m1"),
                        bound=1.0,
                        relative=True,
                        op=ComparisonOp.GEQ,
                    )
                ],
            )
            with self.assertRaisesRegex(
                ValueError, "Relative objective thresholds are not supported"
            ):
                get_hypervolume_trace_of_outcomes_multi_objective(
                    df_wide=df_wide,
                    optimization_config=optimization_config,
                    use_cumulative_hv=True,
                )

        optimization_config = MultiObjectiveOptimizationConfig(
            objective=objective,
        )
        with self.subTest("Cumulative HV"):
            hvs = get_hypervolume_trace_of_outcomes_multi_objective(
                df_wide=df_wide,
                optimization_config=optimization_config,
                use_cumulative_hv=True,
            )
            self.assertEqual(hvs, [0.0, 2.0, 2.0, 3.0])

        with self.subTest("Non-cumulative HV"):
            hvs = get_hypervolume_trace_of_outcomes_multi_objective(
                df_wide=df_wide,
                optimization_config=optimization_config,
                use_cumulative_hv=False,
            )
            self.assertEqual(hvs, [0.0, 2.0, 0.0, 2.0])

    def test_get_trace_by_arm_pull_from_data(self) -> None:
        objective = Objective(metric=Metric("m1"), minimize=False)
        optimzation_config = OptimizationConfig(
            objective=objective,
            outcome_constraints=[
                OutcomeConstraint(
                    Metric("m2"), op=ComparisonOp.GEQ, bound=0.0, relative=False
                )
            ],
        )

        with self.subTest("No data"):
            df = pd.DataFrame()
            result = get_trace_by_arm_pull_from_data(
                df=df, optimization_config=optimzation_config
            )
            self.assertTrue(
                result.equals(
                    pd.DataFrame(columns=["trial_index", "arm_name", "value"])
                )
            )

        with self.subTest("Data for wrong metric"):
            df = pd.DataFrame.from_records(
                data=[
                    {
                        "trial_index": 0,
                        "arm_name": "0_0",
                        "metric_name": "wrong",
                        "mean": 1.0,
                        "sem": None,
                    },
                ]
            )
            with self.assertRaisesRegex(
                ValueError, "Some metrics are not present for all trials and arms"
            ):
                result = get_trace_by_arm_pull_from_data(
                    df=df, optimization_config=optimzation_config
                )

        df = pd.DataFrame.from_records(
            data=[
                {"trial_index": 0, "arm_name": "0_0", "metric_name": "m1", "mean": 1.0},
                {"trial_index": 0, "arm_name": "0_0", "metric_name": "m2", "mean": 1.0},
                # infeasible
                {"trial_index": 0, "arm_name": "0_1", "metric_name": "m1", "mean": 2.0},
                {
                    "trial_index": 0,
                    "arm_name": "0_1",
                    "metric_name": "m2",
                    "mean": -1.0,
                },
                {
                    "trial_index": 1,
                    "arm_name": "0_0",
                    "metric_name": "extraneous",
                    "mean": 4.0,
                },
                {"trial_index": 1, "arm_name": "0_0", "metric_name": "m1", "mean": 3.0},
                {"trial_index": 1, "arm_name": "0_0", "metric_name": "m2", "mean": 0.0},
            ]
        ).assign(sem=None)

        with self.subTest("Relative optimization config not supported"):
            rel_optimization_config = OptimizationConfig(
                objective=objective,
                outcome_constraints=[
                    OutcomeConstraint(
                        metric=Metric("m2"),
                        op=ComparisonOp.GEQ,
                        bound=0.0,
                        relative=True,
                    )
                ],
            )
            with self.assertRaisesRegex(
                ValueError, "Relativized optimization config not supported"
            ):
                get_trace_by_arm_pull_from_data(
                    df=df, optimization_config=rel_optimization_config
                )

        with self.subTest("Single objective, cumulative"):
            result = get_trace_by_arm_pull_from_data(
                df=df, optimization_config=optimzation_config, use_cumulative_best=True
            )
            self.assertEqual(len(result), 3)
            self.assertEqual(set(result.columns), {"trial_index", "arm_name", "value"})
            self.assertEqual(result["value"].tolist(), [1.0, 1.0, 3.0])

        with self.subTest("Single objective, non-cumulative"):
            result = get_trace_by_arm_pull_from_data(
                df=df, optimization_config=optimzation_config, use_cumulative_best=False
            )
            self.assertEqual(len(result), 3)
            self.assertEqual(set(result.columns), {"trial_index", "arm_name", "value"})
            self.assertEqual(result["value"].tolist(), [1.0, float("-inf"), 3.0])

        moo_opt_config = MultiObjectiveOptimizationConfig(
            objective=MultiObjective(
                objectives=[
                    Objective(metric=Metric("m1"), minimize=False),
                    Objective(metric=Metric("m2"), minimize=False),
                ],
            ),
        )
        # reference point inferred to be [1, 0]
        with self.subTest("Multi-objective, cumulative"):
            result = get_trace_by_arm_pull_from_data(
                df=df, optimization_config=moo_opt_config, use_cumulative_best=True
            )
            self.assertEqual(len(result), 3)
            self.assertEqual(set(result.columns), {"trial_index", "arm_name", "value"})
            self.assertEqual(result["value"].tolist(), [0.0, 0.0, 2.0])

        with self.subTest("Multi-objective, non-cumulative"):
            result = get_trace_by_arm_pull_from_data(
                df=df, optimization_config=moo_opt_config, use_cumulative_best=False
            )
            self.assertEqual(len(result), 3)
            self.assertEqual(set(result.columns), {"trial_index", "arm_name", "value"})
            self.assertEqual(result["value"].tolist(), [0.0, 0.0, 2.0])

    @mock_botorch_optimize
    def test_best_from_model_prediction(self) -> None:
        exp = get_branin_experiment()
        gs = choose_generation_strategy_legacy(
            search_space=exp.search_space,
            num_initialization_trials=3,
            suggested_model_override=Generators.BOTORCH_MODULAR,
        )

        for _ in range(3):
            generator_run = gs.gen_single_trial(experiment=exp, n=1)
            trial = exp.new_trial(generator_run=generator_run)
            trial.run().mark_completed()
            exp.attach_data(exp.fetch_data())

        generator_run = gs.gen_single_trial(experiment=exp, n=1)
        trial = exp.new_trial(generator_run=generator_run)
        trial.run().mark_completed()

        with patch.object(
            TorchAdapter,
            "model_best_point",
            return_value=(
                (
                    exp.trials[0].arms[0],
                    (
                        {"branin": 34.76260622783635},
                        {"branin": {"branin": 0.00028306433439807734}},
                    ),
                )
            ),
        ) as mock_model_best_point, self.assertLogs(
            logger=best_point_logger, level="WARN"
        ) as lg:
            # Test bad model fit causes function to resort back to raw data
            with patch(
                "ax.service.utils.best_point.assess_model_fit",
                return_value=AssessModelFitResult(
                    good_fit_metrics_to_fisher_score={},
                    bad_fit_metrics_to_fisher_score={
                        "branin": 0,
                    },
                ),
            ):
                self.assertIsNotNone(
                    get_best_parameters_from_model_predictions_with_trial_index(
                        experiment=exp, adapter=gs.adapter
                    )
                )
                self.assertTrue(
                    any("Model fit is poor" in warning for warning in lg.output),
                    msg=lg.output,
                )
                mock_model_best_point.assert_not_called()

            # Test model best point is used when fit is good
            with patch(
                "ax.service.utils.best_point.assess_model_fit",
                return_value=AssessModelFitResult(
                    good_fit_metrics_to_fisher_score={
                        "branin": 0,
                    },
                    bad_fit_metrics_to_fisher_score={},
                ),
            ):
                self.assertIsNotNone(
                    get_best_parameters_from_model_predictions_with_trial_index(
                        experiment=exp, adapter=gs.adapter
                    )
                )
                mock_model_best_point.assert_called()

        # Assert the non-mocked method works correctly as well
        res = get_best_parameters_from_model_predictions_with_trial_index(
            experiment=exp, adapter=gs.adapter
        )
        trial_index, best_params, predict_arm = none_throws(res)
        self.assertIsNotNone(best_params)
        self.assertIsNotNone(trial_index)
        self.assertIsNotNone(predict_arm)
        # It works even when there are no predictions already stored on the
        # GeneratorRun
        for trial in exp.trials.values():
            trial.generator_run._best_arm_predictions = None
        res = get_best_parameters_from_model_predictions_with_trial_index(
            experiment=exp, adapter=gs.adapter
        )
        trial_index, best_params_no_gr, predict_arm_no_gr = none_throws(res)
        self.assertEqual(best_params, best_params_no_gr)
        self.assertEqual(predict_arm, predict_arm_no_gr)
        self.assertIsNotNone(trial_index)
        self.assertIsNotNone(predict_arm)

    def test_best_raw_objective_point(self) -> None:
        with self.subTest("Only early-stopped trials"):
            exp = get_experiment_with_map_data()
            exp.trials[0].mark_running(no_runner_required=True)
            exp.trials[0].mark_early_stopped(unsafe=True)
            with self.assertRaisesRegex(
                ValueError, "Cannot identify best point if no trials are completed."
            ):
                get_best_raw_objective_point_with_trial_index(experiment=exp)

        exp = get_branin_experiment()
        with self.assertRaisesRegex(
            ValueError, "Cannot identify best point if experiment contains no data."
        ):
            get_best_raw_objective_point_with_trial_index(experiment=exp)
        self.assertIsNone(get_best_by_raw_objective_with_trial_index(exp))
        exp.new_trial(
            generator_run=GeneratorRun(arms=[Arm(parameters={"x1": 5.0, "x2": 5.0})])
        ).run().complete()
        exp.fetch_data()

        with self.subTest("Data present but not for needed metrics"):
            opt_conf = OptimizationConfig(
                objective=Objective(metric=get_branin_metric(name="not_branin"))
            )
            with self.assertRaisesRegex(
                ValueError, "Some metrics are not present for all trials and arms"
            ):
                get_best_raw_objective_point_with_trial_index(
                    experiment=exp, optimization_config=opt_conf
                )

        # Test constraints work as expected.
        observations = [[1.0, 2.0], [3.0, 4.0], [-5.0, -6.0]]
        exp = get_experiment_with_observations(
            observations=observations,
            constrained=True,
            minimize=False,
        )
        _, __, best_prediction = none_throws(
            get_best_by_raw_objective_with_trial_index(exp)
        )
        best_metrics = none_throws(best_prediction)[0]
        self.assertDictEqual(best_metrics, {"m1": 3.0, "m2": 4.0})

        # Tensor bounds are accepted.
        constraint = none_throws(exp.optimization_config).all_constraints[0]
        # pyre-fixme[8]: Attribute `bound` declared in class `OutcomeConstraint`
        # has type `float` but is used as type `Tensor`.
        constraint.bound = torch.tensor(constraint.bound)
        _, __, best_prediction = none_throws(
            get_best_by_raw_objective_with_trial_index(exp)
        )
        best_metrics = none_throws(best_prediction)[0]
        self.assertDictEqual(best_metrics, {"m1": 3.0, "m2": 4.0})

    def test_best_raw_objective_point_unsatisfiable(self) -> None:
        exp = get_branin_experiment()
        trial = exp.new_trial(
            generator_run=GeneratorRun(arms=[Arm(parameters={"x1": 5.0, "x2": 5.0})])
        ).run()
        trial.mark_completed()
        exp.fetch_data()

        opt_conf = none_throws(exp.optimization_config).clone()
        opt_conf.outcome_constraints.append(
            OutcomeConstraint(
                metric=get_branin_metric(), op=ComparisonOp.LEQ, bound=0, relative=False
            )
        )

        with self.assertRaisesRegex(ValueError, "No points satisfied"):
            get_best_raw_objective_point_with_trial_index(
                experiment=exp, optimization_config=opt_conf
            )

        # adding a new metric that is not present in the data should raise an error,
        # even if the other metrics are satisfied
        opt_conf.outcome_constraints.pop()
        unobserved_metric = get_branin_metric(name="unobserved")
        opt_conf.outcome_constraints.append(
            OutcomeConstraint(
                metric=unobserved_metric, op=ComparisonOp.LEQ, bound=0, relative=False
            )
        )
        # also add a constraint that is always satisfied, as the Branin metric is
        # non-negative, and check that only the "unobserved" metric shows up in the
        # error message
        opt_conf.outcome_constraints.append(
            OutcomeConstraint(
                metric=get_branin_metric(), op=ComparisonOp.GEQ, bound=0, relative=False
            )
        )

        with self.assertLogs(logger=best_point_logger, level="WARN") as lg:
            with self.assertRaisesRegex(
                ValueError,
                r"No points satisfied all outcome constraints within 95 percent "
                r"confidence interval\. The feasibility of 1 arm\(s\) could not be "
                r"determined: \['0_0'\]\.",
            ):
                get_best_raw_objective_point_with_trial_index(
                    experiment=exp, optimization_config=opt_conf
                )
        self.assertEqual(len(lg.output), 1)
        self.assertRegex(
            lg.output[0],
            r"Arm 0_0 is missing data for one or more constrained metrics: "
            r"\{'unobserved'\}\.",
        )

    def test_best_raw_objective_point_unsatisfiable_relative(self) -> None:
        exp = get_experiment_with_observations(
            observations=[[-1, 1]],
            constrained=True,
        )

        # Create altered optimization config with unsatisfiable relative constraint.
        opt_conf = none_throws(exp.optimization_config).clone()
        opt_conf.outcome_constraints[0].relative = True
        opt_conf.outcome_constraints[0].bound = 9999

        with self.assertRaisesRegex(
            DataRequiredError,
            "Optimization config has relative constraint, but model was not fit"
            " with status quo.",
        ):
            get_best_raw_objective_point_with_trial_index(
                experiment=exp, optimization_config=opt_conf
            )

        exp.status_quo = exp.trials[0].arms[0]

        with self.assertRaisesRegex(ValueError, "No points satisfied"):
            get_best_raw_objective_point_with_trial_index(exp, opt_conf)

    def test_best_raw_objective_point_scalarized(self) -> None:
        exp = get_branin_experiment()
        gs = choose_generation_strategy_legacy(search_space=exp.search_space)
        exp.optimization_config = OptimizationConfig(
            ScalarizedObjective(metrics=[get_branin_metric()], minimize=True)
        )
        with self.assertRaisesRegex(ValueError, "Cannot identify best "):
            get_best_raw_objective_point_with_trial_index(exp)
        self.assertIsNone(
            get_best_parameters_from_model_predictions_with_trial_index(
                experiment=exp, adapter=gs.adapter
            )
        )
        self.assertIsNone(get_best_by_raw_objective_with_trial_index(experiment=exp))
        params = {"x1": 5.0, "x2": 5.0}
        exp.new_trial(
            generator_run=GeneratorRun(arms=[Arm(parameters=params)])
        ).run().complete()
        exp.fetch_data()
        _, parameterization, __ = get_best_raw_objective_point_with_trial_index(exp)
        self.assertEqual(parameterization, params)

    def test_best_raw_objective_point_scalarized_multi(self) -> None:
        exp = get_branin_experiment()
        gs = choose_generation_strategy_legacy(search_space=exp.search_space)
        exp.optimization_config = OptimizationConfig(
            ScalarizedObjective(
                metrics=[get_branin_metric(), get_branin_metric(lower_is_better=False)],
                weights=[0.1, -0.9],
                minimize=True,
            )
        )
        with self.assertRaisesRegex(ValueError, "Cannot identify best "):
            get_best_raw_objective_point_with_trial_index(experiment=exp)
        self.assertIsNone(
            get_best_parameters_from_model_predictions_with_trial_index(
                experiment=exp, adapter=gs.adapter
            )
        )
        self.assertIsNone(get_best_by_raw_objective_with_trial_index(experiment=exp))
        params = {"x1": 5.0, "x2": 5.0}
        exp.new_trial(
            generator_run=GeneratorRun(arms=[Arm(parameters=params)])
        ).run().complete()
        exp.fetch_data()
        _, parameterization, __ = get_best_raw_objective_point_with_trial_index(exp)
        self.assertEqual(parameterization, params)

    def test_derelativize_opt_config(self) -> None:
        # No change to optimization config without relative constraints/thresholds.
        observations = [[-1, 1, 1], [1, 2, 1], [3, -2, -1], [2, 4, 1], [2, 0, 1]]
        exp = get_experiment_with_observations(
            observations=observations, constrained=True
        )
        input_optimization_config = none_throws(exp.optimization_config)
        with self.subTest("No relative constraints"):
            returned_opt_config = derelativize_opt_config(
                optimization_config=input_optimization_config,
                experiment=exp,
            )
            self.assertEqual(input_optimization_config, returned_opt_config)

        with self.subTest("Provided opt config overrides experiment opt config"):
            opt_config = input_optimization_config.clone_with_args(
                outcome_constraints=[]
            )
            returned_opt_config = derelativize_opt_config(
                optimization_config=opt_config,
                experiment=exp,
            )
            self.assertEqual(opt_config, returned_opt_config)
            self.assertNotEqual(opt_config, exp.optimization_config)

        # Add relative constraints.
        for constraint in input_optimization_config.all_constraints:
            constraint.relative = True

        # Check errors.
        with self.subTest("No fit with status quo"), self.assertRaisesRegex(
            DataRequiredError,
            "Optimization config has relative constraint, but model was not fit"
            " with status quo.",
        ):
            derelativize_opt_config(
                optimization_config=input_optimization_config, experiment=exp
            )

        status_quo_trial_index = 2
        exp.status_quo = exp.trials[status_quo_trial_index].arms[0]
        with self.subTest("Relative constraints and status quo"):
            relativized_opt_config = assert_is_instance(
                derelativize_opt_config(
                    optimization_config=input_optimization_config, experiment=exp
                ),
                MultiObjectiveOptimizationConfig,
            )
            relativized_obj = assert_is_instance(
                relativized_opt_config.objective, MultiObjective
            )
            input_obj = assert_is_instance(
                input_optimization_config.objective, MultiObjective
            )
            status_quo_df = exp.lookup_data_for_trial(
                trial_index=status_quo_trial_index
            )[0].df
            # This is not a real test of `derelativize_opt_config` but rather
            # making sure the values on the experiment have't drifted
            self.assertEqual(status_quo_df["metric_name"].tolist(), ["m1", "m2", "m3"])
            self.assertEqual(status_quo_df["mean"].tolist(), [3.0, -2.0, -1.0])

            status_quo_obs = observations[status_quo_trial_index]
            for i in range(2):
                # Relativization is with respect to the absolute value of the
                # status quo value
                self.assertEqual(
                    relativized_obj.objectives[i].minimize,
                    input_obj.objectives[i].minimize,
                )
                self.assertEqual(
                    relativized_opt_config.objective_thresholds[i].bound,
                    status_quo_obs[i],
                )

            self.assertEqual(
                relativized_opt_config.outcome_constraints[0].bound,
                observations[status_quo_trial_index][2],
            )

    def test_is_row_feasible(self) -> None:
        drop_na_rows = False
        for undetermined_value, drop_na_rows in product([None, False], [True, False]):
            with self.subTest(
                undetermined_value=undetermined_value, drop_na_rows=drop_na_rows
            ):
                self._test_is_row_feasible(
                    undetermined_value=undetermined_value, drop_na_rows=drop_na_rows
                )

    def _test_is_row_feasible(
        self, undetermined_value: bool | None, drop_na_rows: bool
    ) -> None:
        """Tests _is_row_feasible.

        Args:
            undetermined_value: The `undetermined_value` to pass to _is_row_feasible.
            drop_na_rows: Whether to drop the rows corresponding with `NaN` values in
                the dataframe before passing it to `_is_row_feasible`, which tests
                different types of "missing" data (explicit `NaNs` and missing rows).
        """
        exp = get_experiment_with_observations(
            observations=[
                [-1, 1, 1],
                [1, 2, 1],
                [3, 3, -1],
                [2, 4, 1],
                [2, 0, 1],
                # an otherwise feasible observation with a nan observation
                # using 2, 4 because it is feasible for relative constraints too
                [2, 4, np.nan],
                # an otherwise infeasible observation with a nan observation
                [-3, -3, np.nan],
            ],
            constrained=True,
        )

        df = exp.lookup_data().df
        is_na_mask = df["mean"].isna()
        if drop_na_rows:
            df = df[~is_na_mask]

        feasible_series = _is_row_feasible(
            df=df,
            optimization_config=none_throws(exp.optimization_config),
            undetermined_value=undetermined_value,
        )
        # If another metric is already infeasible, the function determines infeasibility
        # even when some other constrained metrics haven't been observed yet.
        expected_per_arm = [False, True, False, True, True, undetermined_value, False]
        expected_series = _repeat_elements(
            list_to_replicate=expected_per_arm, n_repeats=3
        )
        if drop_na_rows:
            expected_series = expected_series[~is_na_mask]

        pd.testing.assert_series_equal(
            feasible_series, expected_series, check_names=False
        )

        exp.optimization_config.outcome_constraints[0].relative = True
        relative_constraint_warning = (
            "WARNING:ax.service.utils.best_point:Determining trial feasibility only "
            "supported with a derelativized OptimizationConfig, but found the "
            "following relative constraints: [OutcomeConstraint(m3 >= 0.0%)]. "
            f"Returning {undetermined_value} as the feasibility."
        )
        with self.assertLogs(logger=best_point_logger, level="WARN") as lg:
            feasible_series = _is_row_feasible(
                df=df,
                optimization_config=none_throws(exp.optimization_config),
                undetermined_value=undetermined_value,
            )
            self.assertTrue(
                any(relative_constraint_warning in warning for warning in lg.output),
                msg=lg.output,
            )
        # When the outcome constraints haven't been derelativized, they are ignored,
        # which leads even the arm with the nan observation to be marked feasible.
        expected_per_arm = [undetermined_value] * len(exp.arms_by_name)
        expected_series = _repeat_elements(
            list_to_replicate=expected_per_arm, n_repeats=3
        )
        if drop_na_rows:
            expected_series = expected_series[~is_na_mask]
        pd.testing.assert_series_equal(
            feasible_series, expected_series, check_names=False
        )
        exp._status_quo = exp.trials[0].arms[0]
        for constraint in none_throws(exp.optimization_config).all_constraints:
            constraint.relative = True
        optimization_config = derelativize_opt_config(
            optimization_config=none_throws(exp.optimization_config),
            experiment=exp,
        )
        with self.assertLogs(logger=best_point_logger, level="WARN") as lg:
            # `assertNoLogs` coming in 3.10 - until then we log a dummy warning and
            # continue.
            best_point_logger.warning("Dummy warning")
            feasible_series = _is_row_feasible(
                df=df,
                optimization_config=optimization_config,
                undetermined_value=undetermined_value,
            )
            self.assertFalse(
                any(relative_constraint_warning in warning for warning in lg.output),
                msg=lg.output,
            )
        expected_per_arm = [True, True, False, True, False, undetermined_value, False]
        expected_series = _repeat_elements(
            list_to_replicate=expected_per_arm, n_repeats=3
        )
        if drop_na_rows:
            expected_series = expected_series[~is_na_mask]
        pd.testing.assert_series_equal(
            feasible_series, expected_series, check_names=False
        )

        # Check that index is carried over for interfacing appropriately
        # with related dataframes.
        exp = get_experiment_with_observations(
            observations=[[-1, 1], [1, 2], [3, 3], [2, 4], [2, 0]],
            constrained=False,
        )
        df = exp.lookup_data().df
        # Artificially redact some data.
        df = df[df["mean"] > 1]
        feasible_series = _is_row_feasible(
            df=df,
            optimization_config=optimization_config,
            undetermined_value=undetermined_value,
        )
        pd.testing.assert_index_equal(
            df.index, feasible_series.index, check_names=False
        )

    def test_compare_to_baseline_select_baseline_name_default_first_trial(self) -> None:
        OBJECTIVE_METRIC = "objective"
        true_obj_metric = Metric(name=OBJECTIVE_METRIC, lower_is_better=True)
        experiment = Experiment(
            search_space=get_branin_search_space(),
            tracking_metrics=[true_obj_metric],
        )

        with patch.object(
            Experiment, "arms_by_name", new_callable=PropertyMock
        ) as mock_arms_by_name:
            mock_arms_by_name.return_value = {"arm1": "value1", "arm2": "value2"}
            self.assertEqual(
                select_baseline_name_default_first_trial(
                    experiment=experiment,
                    baseline_arm_name="arm1",
                ),
                ("arm1", False),
            )

        # specified baseline arm not in trial
        wrong_baseline_name = "wrong_baseline_name"
        with self.assertRaisesRegex(
            ValueError,
            "Arm by name .*" + " not found.",
        ):
            select_baseline_name_default_first_trial(
                experiment=experiment,
                baseline_arm_name=wrong_baseline_name,
            )

        # status quo baseline arm
        experiment_with_status_quo = copy.deepcopy(experiment)
        experiment_with_status_quo.status_quo = Arm(
            name="status_quo",
            parameters={"x1": 0, "x2": 0},
        )
        self.assertEqual(
            select_baseline_name_default_first_trial(
                experiment=experiment_with_status_quo,
                baseline_arm_name=None,
            ),
            ("status_quo", False),
        )
        # first arm from trials
        custom_arm = Arm(name="m_0", parameters={"x1": 0.1, "x2": 0.2})
        experiment.new_trial().add_arm(custom_arm)
        self.assertEqual(
            select_baseline_name_default_first_trial(
                experiment=experiment,
                baseline_arm_name=None,
            ),
            ("m_0", True),
        )

        # none selected
        experiment_with_no_valid_baseline = Experiment(
            search_space=get_branin_search_space(),
            tracking_metrics=[true_obj_metric],
        )

        with self.assertRaisesRegex(
            ValueError,
            "Could not find valid baseline arm.",
        ):
            select_baseline_name_default_first_trial(
                experiment=experiment_with_no_valid_baseline,
                baseline_arm_name=None,
            )

    # NOTE: Can't use mock optimize here, since we rely on model predictions.
    # It is still very cheap even with model fitting.
    def test_get_best_point_with_model_prediction(
        self,
    ) -> None:
        ax_client = AxClient()
        ax_client.create_experiment(
            name="test_experiment",
            parameters=[
                {
                    "name": "x",
                    "type": "range",
                    "bounds": [1.0, 10.0],
                },
            ],
            objectives={"y": ObjectiveProperties(minimize=True)},
            is_test=True,
            choose_generation_strategy_kwargs={"num_initialization_trials": 2},
        )

        params, idx = ax_client.get_next_trial()
        ax_client.complete_trial(idx, raw_data={"y": 0})

        for i in range(1, 3):
            ax_client.get_next_trial()
            ax_client.complete_trial(i, raw_data={"y": i})

        # Mock with no bad fir metrics ensures that the model is used
        # to extract the best point.
        with patch(
            f"{best_point_module}.assess_model_fit",
            return_value=AssessModelFitResult(
                good_fit_metrics_to_fisher_score={"y": 1},
                bad_fit_metrics_to_fisher_score={},
            ),
        ) as mock_model_fit:
            best_index, best_params, predictions = none_throws(
                ax_client.get_best_trial()
            )
        mock_model_fit.assert_called_once()
        self.assertEqual(best_index, idx)
        self.assertEqual(best_params, params)
        # We should get both mean & covariance predictions.
        self.assertEqual(predictions, ({"y": mock.ANY}, {"y": {"y": mock.ANY}}))

        # Also verify that fallback option from GR produces the right trial index.
        gr = ax_client.experiment.trials[2].generator_runs[0]
        best_index, best_params, predictions = none_throws(
            _extract_best_arm_from_gr(gr=gr, trials=ax_client.experiment.trials)
        )
        self.assertEqual(best_index, idx)
        self.assertEqual(best_params, params)
        self.assertEqual(predictions, ({"y": mock.ANY}, {"y": {"y": mock.ANY}}))


def _repeat_elements(list_to_replicate: list[Any], n_repeats: int) -> pd.Series:
    return pd.Series([item for item in list_to_replicate for _ in range(n_repeats)])



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest.mock import patch

from ax.core.base_trial import BaseTrial, TrialStatus
from ax.core.multi_type_experiment import (
    filter_trials_by_type,
    get_trial_indices_for_statuses,
)
from ax.core.objective import Objective
from ax.core.optimization_config import OptimizationConfig
from ax.metrics.branin import BraninMetric
from ax.runners.synthetic import SyntheticRunner
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_arms, get_multi_type_experiment


class MultiTypeExperimentTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.experiment = get_multi_type_experiment()

    def test_MTExperimentFlow(self) -> None:
        self.assertTrue(self.experiment.supports_trial_type("type1"))
        self.assertTrue(self.experiment.supports_trial_type("type2"))
        self.assertFalse(self.experiment.supports_trial_type(None))

        n = 10
        arms = get_branin_arms(n=n, seed=0)

        b1 = self.experiment.new_batch_trial()
        b1.add_arms_and_weights(arms=arms)
        self.assertEqual(b1.trial_type, "type1")
        b1.run()
        self.assertEqual(b1.run_metadata["dummy_metadata"], "dummy1")

        self.experiment.update_runner("type2", SyntheticRunner(dummy_metadata="dummy3"))
        b2 = self.experiment.new_batch_trial()
        b2.trial_type = "type2"
        b2.add_arms_and_weights(arms=arms)
        self.assertEqual(b2.trial_type, "type2")
        b2.run()
        self.assertEqual(b2.run_metadata["dummy_metadata"], "dummy3")

        df = self.experiment.fetch_data().df
        for _, row in df.iterrows():
            # Make sure proper metric present for each batch only
            self.assertEqual(
                row["metric_name"], "m1" if row["trial_index"] == 0 else "m2"
            )

        arm_0_slice = df.loc[df["arm_name"] == "0_0"]
        self.assertNotEqual(
            float(arm_0_slice[df["trial_index"] == 0]["mean"].item()),
            float(arm_0_slice[df["trial_index"] == 1]["mean"].item()),
        )
        self.assertEqual(len(df), 2 * n)
        self.assertEqual(self.experiment.default_trials, {0})
        # Set 2 metrics to be equal
        self.experiment.update_tracking_metric(
            BraninMetric("m2", ["x1", "x2"]), trial_type="type2"
        )
        df = self.experiment.fetch_data().df
        arm_0_slice = df.loc[df["arm_name"] == "0_0"]
        self.assertAlmostEqual(
            float(arm_0_slice[df["trial_index"] == 0]["mean"].item()),
            float(arm_0_slice[df["trial_index"] == 1]["mean"].item()),
            places=10,
        )

    def test_Repr(self) -> None:
        self.assertEqual(str(self.experiment), "MultiTypeExperiment(test_exp)")

    def test_Eq(self) -> None:
        exp2 = get_multi_type_experiment()

        # Should be equal to start
        self.assertTrue(self.experiment == exp2)

        self.experiment.add_tracking_metric(
            BraninMetric("m3", ["x2", "x1"]), trial_type="type1", canonical_name="m4"
        )

        # Test different set of metrics
        self.assertFalse(self.experiment == exp2)

        exp2.add_tracking_metric(
            BraninMetric("m3", ["x2", "x1"]), trial_type="type1", canonical_name="m5"
        )

        # Test different metric definitions
        self.assertFalse(self.experiment == exp2)

        exp2.update_tracking_metric(
            BraninMetric("m3", ["x2", "x1"]), trial_type="type1", canonical_name="m4"
        )

        # Should be the same
        self.assertTrue(self.experiment == exp2)

        exp2.remove_tracking_metric("m3")
        self.assertFalse(self.experiment == exp2)

    def test_BadBehavior(self) -> None:
        # Add trial type that already exists
        with self.assertRaises(ValueError):
            self.experiment.add_trial_type("type1", SyntheticRunner())

        # Update runner for non-existent trial type
        with self.assertRaises(ValueError):
            self.experiment.update_runner("type3", SyntheticRunner())

        # Add metric for trial_type that doesn't exist
        with self.assertRaises(ValueError):
            self.experiment.add_tracking_metric(
                BraninMetric("m2", ["x1", "x2"]), "type3"
            )

        # Try to remove metric that doesn't exist
        with self.assertRaises(ValueError):
            self.experiment.remove_tracking_metric("m3")

        # Try to change optimization metric to non-primary trial type
        with self.assertRaises(ValueError):
            self.experiment.update_tracking_metric(
                BraninMetric("m1", ["x1", "x2"]), "type2"
            )

        # Update metric definition for trial_type that doesn't exist
        with self.assertRaises(ValueError):
            self.experiment.update_tracking_metric(
                BraninMetric("m2", ["x1", "x2"]), "type3"
            )

        # Try to get runner for trial_type that's not supported
        batch = self.experiment.new_batch_trial()
        batch._trial_type = "type3"  # Force override trial type
        with self.assertRaises(ValueError):
            self.experiment.runner_for_trial(batch)

        # Try making trial with unsupported trial type
        with self.assertRaises(ValueError):
            self.experiment.new_batch_trial(trial_type="type3")

        # Try resetting runners.
        with self.assertRaises(NotImplementedError):
            self.experiment.reset_runners(SyntheticRunner())

    def test_setting_opt_config(self) -> None:
        self.assertDictEqual(
            self.experiment._metric_to_trial_type, {"m1": "type1", "m2": "type2"}
        )
        self.experiment.optimization_config = OptimizationConfig(
            Objective(BraninMetric("m3", ["x1", "x2"]), minimize=True)
        )
        self.assertDictEqual(
            self.experiment._metric_to_trial_type,
            {"m1": "type1", "m2": "type2", "m3": "type1"},
        )

    def test_runner_for_trial_type(self) -> None:
        runner = self.experiment.runner_for_trial_type(trial_type="type1")
        self.assertIs(runner, self.experiment._trial_type_to_runner["type1"])
        with self.assertRaisesRegex(
            ValueError, "Trial type `invalid` is not supported."
        ):
            self.experiment.runner_for_trial_type(trial_type="invalid")

    def test_add_tracking_metrics(self) -> None:
        type1_metrics = [
            BraninMetric("m3_type1", ["x1", "x2"]),
            BraninMetric("m4_type1", ["x1", "x2"]),
        ]
        type2_metrics = [
            BraninMetric("m3_type2", ["x1", "x2"]),
            BraninMetric("m4_type2", ["x1", "x2"]),
        ]
        default_type_metrics = [
            BraninMetric("m5_default_type", ["x1", "x2"]),
        ]
        self.experiment.add_tracking_metrics(
            metrics=type1_metrics + type2_metrics + default_type_metrics,
            metrics_to_trial_types={
                "m3_type1": "type1",
                "m4_type1": "type1",
                "m3_type2": "type2",
                "m4_type2": "type2",
            },
        )
        self.assertDictEqual(
            self.experiment._metric_to_trial_type,
            {
                "m1": "type1",
                "m2": "type2",
                "m3_type1": "type1",
                "m4_type1": "type1",
                "m3_type2": "type2",
                "m4_type2": "type2",
                "m5_default_type": "type1",
            },
        )

    def test_stop_trial_runs_multi_type_experiment(self) -> None:
        # Setup 3 trials with 2 runners
        self.experiment.new_batch_trial(trial_type="type1")
        self.experiment.new_batch_trial(trial_type="type2")
        self.experiment.new_batch_trial(trial_type="type2")
        runner1 = self.experiment.runner_for_trial_type(trial_type="type1")
        runner2 = self.experiment.runner_for_trial_type(trial_type="type2")

        # Mock is needed because SyntheticRunner does not implement a 'stop'
        # method
        with patch.object(
            runner1, "stop", return_value=None
        ) as mock_runner_stop1, patch.object(
            runner2, "stop", return_value=None
        ) as mock_runner_stop2, patch.object(
            BaseTrial, "mark_early_stopped"
        ) as mock_mark_stopped:
            self.experiment.stop_trial_runs(
                trials=[self.experiment.trials[0], self.experiment.trials[1]]
            )
        mock_runner_stop1.assert_called_once()
        mock_runner_stop2.assert_called()
        mock_mark_stopped.assert_called()


class MultiTypeExperimentUtilsTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.experiment = get_multi_type_experiment()
        self.experiment.new_batch_trial(trial_type="type1")
        self.experiment.new_batch_trial(trial_type="type2")

    def test_filter_trials_by_type(self) -> None:
        trials = self.experiment.trials.values()
        self.assertEqual(len(trials), 2)
        filtered = filter_trials_by_type(trials, trial_type="type1")
        self.assertEqual(len(filtered), 1)
        self.assertEqual(filtered[0].trial_type, "type1")
        filtered = filter_trials_by_type(trials, trial_type="type2")
        self.assertEqual(len(filtered), 1)
        self.assertEqual(filtered[0].trial_type, "type2")
        filtered = filter_trials_by_type(trials, trial_type="invalid")
        self.assertEqual(len(filtered), 0)
        filtered = filter_trials_by_type(trials, trial_type=None)
        self.assertEqual(len(filtered), 2)

    def test_get_trial_indices_for_statuses(self) -> None:
        self.assertEqual(
            get_trial_indices_for_statuses(
                experiment=self.experiment,
                statuses={TrialStatus.CANDIDATE, TrialStatus.STAGED},
                trial_type="type1",
            ),
            {0},
        )
        self.assertEqual(
            get_trial_indices_for_statuses(
                experiment=self.experiment,
                statuses={TrialStatus.CANDIDATE, TrialStatus.STAGED},
                trial_type="type2",
            ),
            {1},
        )
        self.assertEqual(
            get_trial_indices_for_statuses(
                experiment=self.experiment,
                statuses={TrialStatus.CANDIDATE, TrialStatus.STAGED},
            ),
            {0, 1},
        )
        self.experiment.trials[0].mark_running(no_runner_required=True)
        self.experiment.trials[1].mark_abandoned()
        self.assertEqual(
            get_trial_indices_for_statuses(
                experiment=self.experiment,
                statuses={TrialStatus.RUNNING},
                trial_type="type1",
            ),
            {0},
        )
        self.assertEqual(
            get_trial_indices_for_statuses(
                experiment=self.experiment,
                statuses={TrialStatus.ABANDONED},
                trial_type="type2",
            ),
            {1},
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Optional, TYPE_CHECKING

from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.base import Transform
from ax.core.observation import Observation, ObservationFeatures, separate_observations
from ax.core.parameter import PARAMETER_PYTHON_TYPE_MAP, RangeParameter
from ax.core.search_space import HierarchicalSearchSpace, SearchSpace
from ax.exceptions.core import UserInputError
from ax.generators.types import TConfig
from pandas import Series
from pyre_extensions import assert_is_instance, none_throws

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


class Cast(Transform):
    """Cast each param value to the respective parameter's type/format and
    to a flattened version of the hierarchical search space, if applicable.

    This is a default transform that should run across all models.

    NOTE: In case where the search space is hierarchical and this transform is
    configured to flatten it:
      * All calls to `Cast.transform_...` transform Ax objects defined in
        terms of hierarchical search space, to their definitions in terms of
        flattened search space.
      * All calls to `Cast.untransform_...` cast Ax objects back to a
        hierarchical search space.
      * The hierarchical search space is seen as the "original" search space,
        and the flattened search space---as "transformed".

    Transform is done in-place for casting types, but objects are copied
    during flattening of- and casting to the hierarchical search space.
    """

    def __init__(
        self,
        search_space: SearchSpace | None = None,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: Optional["adapter_module.base.Adapter"] = None,
        config: TConfig | None = None,
    ) -> None:
        self.search_space: SearchSpace = none_throws(search_space).clone()
        super().__init__(
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )
        config = (config or {}).copy()

        # 1. The default behavior for non-hierarchical search spaces: The flag
        # `flatten_hss` is irrelevant. It's simply ignored.
        #
        # 2. The default behavior for hierarchical search spaces: `flatten_hss=True`.
        # The search space is flattened, and the inactive parameters are filled with
        # dummy values.
        #
        # 3. To exploit the hierarchical structure, one need set `flatten_hss=False`.
        # The search space will be kept as `HierarchicalSearchSpace`. The observations
        # features are still flattened so that the resulting tensors received by BoTorch
        # have the same dimensionality. If the underlying BoTorch model is designed to
        # exploit the hierarchical structure, it infers which parameter is active based
        # on `search_space_digest.hierarchical_dependencies`.
        self.flatten_hss: bool = assert_is_instance(
            config.pop(
                "flatten_hss", isinstance(search_space, HierarchicalSearchSpace)
            ),
            bool,
        )
        self.inject_dummy_values_to_complete_flat_parameterization: bool = (
            assert_is_instance(
                config.pop(
                    "inject_dummy_values_to_complete_flat_parameterization", True
                ),
                bool,
            )
        )
        self.use_random_dummy_values: bool = assert_is_instance(
            config.pop("use_random_dummy_values", False), bool
        )
        if config:
            raise UserInputError(
                f"Unexpected config parameters for `Cast` transform: {config}."
            )

    def _transform_search_space(self, search_space: SearchSpace) -> SearchSpace:
        """Flattens the hierarchical search space and returns the flat
        ``SearchSpace`` if this transform is configured to flatten hierarchical
        search spaces. Does nothing if the search space is not hierarchical.

        NOTE: All calls to `Cast.transform_...` transform Ax objects defined in
        terms of hierarchical search space, to their definitions in terms of
        flattened search space. All calls to `Cast.untransform_...` cast Ax
        objects back to a hierarchical search space.

        Args:
            search_space: The search space to flatten.

        Returns: transformed search space.
        """
        if isinstance(search_space, HierarchicalSearchSpace):
            return search_space.flatten() if self.flatten_hss else search_space
        else:
            return search_space

    def transform_observations(
        self, observations: list[Observation]
    ) -> list[Observation]:
        """Transform observations.

        Typically done in place. By default, the effort is split into separate
        transformations of the features and the data.

        NOTE: We overwrite it here, since ``transform_observation_features`` will drop
        features with ``None`` in them, leading to errors in the base implementation.

        Args:
            observations: Observations.

        Returns: transformed observations.
        """
        obs_feats, obs_data = separate_observations(observations=observations)
        # NOTE: looping here is ok, since the underlying methods for Cast also process
        # the features one by one in a loop.
        trans_obs = []
        for obs_ft, obs_d, obs in zip(obs_feats, obs_data, observations, strict=True):
            tf_obs_feats = self.transform_observation_features(
                observation_features=[obs_ft]
            )
            if len(tf_obs_feats) == 1:
                # Only re-package if the observation features haven't been dropped.
                trans_obs.append(
                    Observation(
                        features=tf_obs_feats[0], data=obs_d, arm_name=obs.arm_name
                    )
                )

        return trans_obs

    def transform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        """Transform observation features by
        - adding parameter values that were removed during casting of observation
          features to hierarchical search space;
        - casting parameter values to the corresponding parameter type;
        - dropping any observations with ``None`` parameter values.

        Args:
            observation_features: Observation features

        Returns: transformed observation features
        """
        observation_features = self._cast_parameter_values(
            observation_features=observation_features
        )

        if isinstance(self.search_space, HierarchicalSearchSpace):
            # Inject the parameters model suggested in the flat search space, which then
            # got removed during casting to HSS as they were not applicable under the
            # hierarchical structure of the search space.
            return [
                # pyre-ignore[16]: The search space in this if-branch is a HSS, and
                # does have `flatten_observation_features`.
                self.search_space.flatten_observation_features(
                    observation_features=obs_feats,
                    inject_dummy_values_to_complete_flat_parameterization=(
                        self.inject_dummy_values_to_complete_flat_parameterization
                    ),
                    use_random_dummy_values=self.use_random_dummy_values,
                )
                for obs_feats in observation_features
            ]
        else:
            return observation_features

    def untransform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        """Untransform observation features by casting parameter values to their
        expected types and removing parameter values that are not applicable given
        the values of other parameters and the hierarchical structure of the search
        space.

        Args:
            observation_features: Observation features in the transformed space

        Returns: observation features in the original space
        """
        observation_features = self._cast_parameter_values(
            observation_features=observation_features
        )

        if isinstance(self.search_space, HierarchicalSearchSpace):
            # The inactive parameters in the HSS have been filled with dummy values,
            # which should be removed from the observations.
            return [
                # pyre-ignore[16]: The search space in this if-branch is a HSS, and
                # does have `case_observation_features`.
                self.search_space.cast_observation_features(
                    observation_features=obs_feats
                )
                for obs_feats in observation_features
            ]
        else:
            return observation_features

    def _cast_parameter_values(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        """Cast parameter values of the given ``ObseravationFeatures`` to the
        ``ParameterType`` of the corresponding parameters in the search space.

        NOTE: This is done in-place. ``ObservationFeatures`` with ``None``
        values are dropped.

        Args:
            observation_features: A list of ``ObservationFeatures`` to cast.

        Returns: observation features with casted parameter values.
        """
        new_obsf = []
        for obsf in observation_features:
            for p_name, p_value in obsf.parameters.items():
                if p_value is None:
                    # Skip obsf if there are `None`s.
                    # The else block below will not be executed.
                    break
                if p_name in self.search_space.parameters:
                    obsf.parameters[p_name] = self.search_space[p_name].cast(p_value)
            else:
                # No `None`s in the parameterization.
                new_obsf.append(obsf)
        return new_obsf

    def transform_experiment_data(
        self, experiment_data: ExperimentData
    ) -> ExperimentData:
        """Cast the column values of the given experiment data to the
        data type corresponding to the ``ParameterType``. If any rows
        include None / NaN values for the parameterization, they are dropped.

        For `HierarchicalSearchSpace`, the parameterizations are flattened
        before applying the rest of the transform.
        """
        arm_data = experiment_data.arm_data
        # If applicable, flatten first to fill otherwise NaN values.
        # See `CastTransformTest.test_transform_experiment_data_flatten` for an
        # example of the arm_data before and after the flattening operation.
        # Any parameter values that are missing in the corresponding columns
        # are filled with the values from `metadata["Keys.FULL_PARAMETERIZATION"]`.
        # If the metadata does not include the full parameterization, a dummy
        # value is constructed, either randomly or by using the middle of the
        # parameter domain, depending on the `use_random_dummy_values` flag.
        if isinstance(self.search_space, HierarchicalSearchSpace):
            # NOTE: This could probably be vectorized to operate more efficiently,
            # however it is non-trivial since we need to extract values from the
            # metadata of each row and fill any missing values after.
            # This is applying the logic of `Cast.transform_observation_features`
            # to the df as a quick solution.
            def flatten_row(row: Series) -> Series:
                obs_ft = ObservationFeatures(
                    parameters=row.drop("metadata").dropna().to_dict(),
                    metadata=row["metadata"],
                )
                flattened = assert_is_instance(
                    self.search_space, HierarchicalSearchSpace
                ).flatten_observation_features(
                    observation_features=obs_ft,
                    inject_dummy_values_to_complete_flat_parameterization=(
                        self.inject_dummy_values_to_complete_flat_parameterization
                    ),
                    use_random_dummy_values=self.use_random_dummy_values,
                )
                return Series(
                    name=row.name,
                    data=flattened.parameters | {"metadata": row["metadata"]},
                )

            arm_data = arm_data.apply(flatten_row, axis=1)

        parameter_names = list(arm_data.columns)
        parameter_names.remove("metadata")

        # Drop rows with None / NaN values. These are dropped in `parameter.cast`
        # call in `Cast.transform_observation_features`.
        arm_data = arm_data.dropna(axis="index", how="any", subset=parameter_names)
        # Update observation_data to include the same indices.
        observation_data = experiment_data.observation_data
        index_names = list(observation_data.index.names)
        if len(index_names) == 2:
            observation_data = observation_data.loc[arm_data.index]
        else:
            observation_data = observation_data.loc[
                observation_data.index.droplevel(level=index_names[2:]).isin(
                    arm_data.index
                )
            ]
        # Add any missing columns to the arm_data to complete it.
        arm_data = arm_data.reindex(
            columns=list(self.search_space.parameters) + ["metadata"], fill_value=None
        )
        # Cast columns to the correct datatype & round RangeParameters, if applicable.
        column_to_type = {
            p: PARAMETER_PYTHON_TYPE_MAP[param.parameter_type]
            for p, param in self.search_space.parameters.items()
        }
        arm_data = arm_data.astype(dtype=column_to_type, copy=False)
        # Round to digits if any parameter specifies it.
        for p_name in parameter_names:
            parameter = self.search_space.parameters[p_name]
            if isinstance(parameter, RangeParameter) and parameter.digits is not None:
                arm_data[p_name] = arm_data[p_name].round(parameter.digits)

        return ExperimentData(arm_data=arm_data, observation_data=observation_data)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict
from ax.adapter.base import Adapter
from ax.analysis.analysis import Analysis
from ax.analysis.analysis_card import AnalysisCardGroup
from ax.analysis.plotly.arm_effects import ArmEffectsPlot
from ax.analysis.utils import extract_relevant_adapter
from ax.core.base_trial import BaseTrial, TrialStatus
from ax.core.batch_trial import BatchTrial
from ax.core.experiment import Experiment
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from pyre_extensions import override

TRIALS_CARDGROUP_TITLE = "Trial-Level Analyses"
TRIALS_CARDGROUP_SUBTITLE = (
    "This analysis provides detailed information about the individual trials in the "
    "experiment. It contains visualizations specifically computed for each trial."
)


class AllTrialsAnalysis(Analysis):
    """
    An Analysis that provides detailed information about all trials in an experiment.

    AllTrialsAnalysis serves as a container for trial-level analyses, organizing them
    into separate card groups. Each child in the card group represents the output of
    TrialAnalysis for a specific trial in the experiment.
    """

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCardGroup:
        # Ensure we have an Experiment provided by the user to extract the relevant
        # metric names from.
        if experiment is None:
            raise UserInputError("TrialsAnalysis requires an Experiment.")

        trials = experiment.trials.values()

        return AnalysisCardGroup(
            name="AllTrialsAnalysis",
            title=TRIALS_CARDGROUP_TITLE,
            subtitle=TRIALS_CARDGROUP_SUBTITLE,
            children=[
                TrialAnalysis(trial=trial).compute_or_error_card(
                    experiment=experiment,
                    generation_strategy=generation_strategy,
                    adapter=adapter,
                )
                for trial in trials
                if isinstance(trial, BatchTrial)
            ],
        )


class TrialAnalysis(Analysis):
    """
    An Analysis that provides detailed information about a specific trial.

    TrialAnalysis computes and organizes various analyses specific to a single trial,
    such as predicted effects plots.
    """

    @override
    def __init__(
        self,
        trial: BaseTrial,
    ) -> None:
        self.trial = trial

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCardGroup:
        # Ensure we have an Experiment provided by the user to extract the relevant
        # metric names from.
        if experiment is None:
            raise UserInputError("TrialAnalysis requires an Experiment.")

        relevant_adapter = extract_relevant_adapter(
            experiment=experiment,
            generation_strategy=generation_strategy,
            adapter=adapter,
        )

        # If the Experiment has an OptimizationConfig set, extract the objective and
        # constraint names.
        objective_names = []
        constraint_names = []
        if (optimization_config := experiment.optimization_config) is not None:
            objective_names = optimization_config.objective.metric_names
            constraint_names = [
                constraint.metric.name
                for constraint in optimization_config.outcome_constraints
            ]

        relativize = experiment.status_quo is not None and isinstance(
            self.trial, BatchTrial
        )
        analyses = [
            ArmEffectsPlot(
                metric_names=[*objective_names, *constraint_names],
                use_model_predictions=True,
                relativize=relativize,
                trial_index=self.trial.index,
            )
            if self.trial.status == TrialStatus.CANDIDATE
            else None
        ]
        return AnalysisCardGroup(
            name=f"{self.trial.index}",
            title=f"Trial {self.trial.index}",
            subtitle=(
                "Trial-level visualizations computed specifically for this trial."
            ),
            children=[
                analysis.compute_or_error_card(
                    experiment=experiment,
                    generation_strategy=generation_strategy,
                    adapter=relevant_adapter,
                )
                for analysis in analyses
                if analysis is not None
            ],
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from copy import deepcopy
from functools import partial
from itertools import product

import numpy as np

from ax.adapter.base import Adapter, DataLoaderConfig
from ax.adapter.data_utils import extract_experiment_data
from ax.adapter.transforms.bilog_y import bilog_transform, BilogY, inv_bilog_transform
from ax.adapter.transforms.log_y import match_ci_width
from ax.core.observation import observations_from_data
from ax.generators.base import Generator
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment
from pandas.testing import assert_frame_equal, assert_series_equal


class BilogYTest(TestCase):
    def setUp(self) -> None:
        self.exp = get_branin_experiment(
            with_status_quo=True,
            with_completed_batch=True,
            with_absolute_constraint=True,
            with_relative_constraint=True,
        )
        self.data = self.exp.fetch_data()
        self.bound = self.exp.optimization_config.outcome_constraints[1].bound

    def get_adapter(self) -> Adapter:
        return Adapter(
            search_space=self.exp.search_space,
            generator=Generator(),
            experiment=self.exp,
            data=self.exp.lookup_data(),
        )

    def test_Init(self) -> None:
        # With adapter.
        t = BilogY(
            search_space=self.exp.search_space,
            adapter=self.get_adapter(),
        )
        self.assertEqual(t.metric_to_bound, {"branin_e": self.bound})

        with self.subTest("With no adapter"):
            t = BilogY(
                search_space=self.exp.search_space,
                adapter=None,
            )
            self.assertEqual(t.metric_to_bound, {})

    def test_Bilog(self) -> None:
        self.assertAlmostEqual(
            float(bilog_transform(y=np.array(7.3), bound=3)), 4.667706820558076
        )
        self.assertAlmostEqual(
            float(bilog_transform(y=np.array(7.3), bound=10)), 8.691667180349821
        )
        self.assertAlmostEqual(
            float(inv_bilog_transform(y=np.array(4.3), bound=3)), 5.669296667619244
        )
        self.assertAlmostEqual(
            float(inv_bilog_transform(y=np.array(2.3), bound=3)), 1.9862472925295231
        )
        self.assertAlmostEqual(
            float(
                inv_bilog_transform(bilog_transform(y=np.array(0.3), bound=3), bound=3)
            ),
            0.3,
        )
        self.assertAlmostEqual(
            float(
                bilog_transform(inv_bilog_transform(y=np.array(0.3), bound=3), bound=3)
            ),
            0.3,
        )

    def test_TransformUntransform(self) -> None:
        bound = self.exp.optimization_config.outcome_constraints[0].bound
        observations = observations_from_data(
            experiment=self.exp, data=self.exp.lookup_data()
        )
        t = BilogY(
            search_space=self.exp.search_space,
            adapter=self.get_adapter(),
        )

        # Transform
        transformed_data = t._transform_observation_data(
            deepcopy([obs.data for obs in observations])
        )
        for obs, transform_obs in zip(observations, transformed_data):
            # Non-constraints should be the same
            self.assertEqual(
                transform_obs.metric_names, ["branin", "branin_d", "branin_e"]
            )
            self.assertTrue((transform_obs.means[0:2] == obs.data.means[0:2]).all())
            self.assertTrue(
                (
                    transform_obs.covariance[0:2, 0:2] == obs.data.covariance[0:2, 0:2]
                ).all()
            )
            # Make sure the bilog transform with ci width matching was applied
            self.assertAlmostEqual(
                transform_obs.means[2], bilog_transform(obs.data.means[2], bound=bound)
            )
            self.assertTrue(  # The transformed variance should be smaller
                transform_obs.covariance[2, 2] < obs.data.covariance[2, 2]
            )

        # Untransform
        untransformed_data = t._untransform_observation_data(deepcopy(transformed_data))
        for obs, untransform_obs, transform_obs in zip(
            observations, untransformed_data, transformed_data
        ):
            # Non-constraints should be the same
            self.assertEqual(
                untransform_obs.metric_names, ["branin", "branin_d", "branin_e"]
            )
            self.assertTrue((untransform_obs.means[0:2] == obs.data.means[0:2]).all())
            self.assertTrue(
                (
                    untransform_obs.covariance[0:2, 0:2]
                    == obs.data.covariance[0:2, 0:2]
                ).all()
            )
            # Make sure the inverse bilog transform with ci width matching was applied
            self.assertAlmostEqual(
                untransform_obs.means[2],
                inv_bilog_transform(transform_obs.means[2], bound=bound),
            )
            # And that we are back where we started as invf(f(x)) = x
            self.assertAlmostEqual(
                obs.data.means[2],
                inv_bilog_transform(transform_obs.means[2], bound=bound),
            )
            self.assertAlmostEqual(
                untransform_obs.covariance[2, 2], obs.data.covariance[2, 2]
            )

    def test_TransformOptimizationConfig(self) -> None:
        t = BilogY(
            search_space=self.exp.search_space,
            adapter=self.get_adapter(),
        )
        oc = self.exp.optimization_config
        # This should be a no-op
        new_oc = t.transform_optimization_config(optimization_config=oc)
        self.assertEqual(new_oc, oc)

    def test_TransformSearchSpace(self) -> None:
        t = BilogY(search_space=self.exp.search_space, adapter=self.get_adapter())
        # This should be a no-op
        new_ss = t.transform_search_space(self.exp.search_space)
        self.assertEqual(new_ss, self.exp.search_space)

    def test_transform_experiment_data(self) -> None:
        t = BilogY(search_space=self.exp.search_space, adapter=self.get_adapter())
        experiment_data = extract_experiment_data(
            experiment=self.exp, data_loader_config=DataLoaderConfig()
        )
        transformed_data = t.transform_experiment_data(
            experiment_data=deepcopy(experiment_data)
        )

        # Check that arm data is identical.
        assert_frame_equal(transformed_data.arm_data, experiment_data.arm_data)

        # Check that non-constraint metrics are unchanged.
        cols = list(product(("mean", "sem"), ("branin", "branin_d")))
        assert_frame_equal(
            transformed_data.observation_data[cols],
            experiment_data.observation_data[cols],
        )

        # Check that `branin_e` has been transformed correctly.
        assert_series_equal(
            transformed_data.observation_data[("mean", "branin_e")],
            bilog_transform(
                experiment_data.observation_data[("mean", "branin_e")], bound=self.bound
            ),
        )
        # Sem is smaller than before.
        self.assertTrue(
            (
                transformed_data.observation_data[("sem", "branin_e")]
                < experiment_data.observation_data[("sem", "branin_e")]
            ).all()
        )
        # Compare against transforming the old way.
        mean, var = match_ci_width(
            mean=experiment_data.observation_data[("mean", "branin_e")],
            sem=None,
            variance=experiment_data.observation_data[("sem", "branin_e")] ** 2,
            transform=partial(bilog_transform, bound=self.bound),
        )
        assert_series_equal(
            transformed_data.observation_data[("mean", "branin_e")], mean
        )
        # Can't use assert_series_equal since the metadata is destroyed in var.
        self.assertTrue(
            transformed_data.observation_data[("sem", "branin_e")].equals(var**0.5)
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import math

import pandas as pd
from ax.adapter.base import Adapter
from ax.analysis.analysis import Analysis
from ax.analysis.analysis_card import AnalysisCardBase
from ax.analysis.plotly.color_constants import AX_BLUE

from ax.analysis.plotly.plotly_analysis import (
    create_plotly_analysis_card,
    PlotlyAnalysisCard,
)
from ax.analysis.plotly.surface.utils import (
    get_parameter_values,
    is_axis_log_scale,
    select_fixed_value,
)
from ax.analysis.plotly.utils import (
    get_scatter_point_color,
    select_metric,
    truncate_label,
    Z_SCORE_95_CI,
)
from ax.analysis.utils import extract_relevant_adapter, relativize_data
from ax.core.experiment import Experiment
from ax.core.observation import ObservationFeatures
from ax.core.trial_status import STATUSES_EXPECTING_DATA
from ax.core.utils import get_target_trial_index
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from plotly import graph_objects as go
from pyre_extensions import none_throws, override

SLICE_CARDGROUP_TITLE = "Slice Plots: Metric effects by parameter value"

SLICE_CARDGROUP_SUBTITLE = (
    "These plots show the relationship between a metric and a parameter. They "
    "show the predicted values of the metric on the y-axis as a function of the "
    "parameter on the x-axis while keeping all other parameters fixed at their "
    "status_quo value (or mean value if status_quo is unavailable). "
)


class SlicePlot(Analysis):
    """
    Plot a 1D "slice" of the surrogate model's predicted outcomes for a given
    parameter, where all other parameters are held fixed at their status-quo value or
    mean if no status quo is available.

    The DataFrame computed will contain the following columns:
        - PARAMETER_NAME: The value of the parameter specified
        - METRIC_NAME_mean: The predected mean of the metric specified
        - METRIC_NAME_sem: The predected sem of the metric specified
        - sampled: Whether the parameter value was sampled in at least one trial
    """

    def __init__(
        self,
        parameter_name: str,
        metric_name: str | None = None,
        display_sampled: bool = True,
        relativize: bool = False,
    ) -> None:
        """
        Args:
            parameter_name: The name of the parameter to plot on the x axis.
            metric_name: The name of the metric to plot on the y axis. If not
                specified the objective will be used.
            display_sampled: If True, plot "x"s at x coordinates which have been
                sampled in at least one trial.
            relativize: If True, relativize the metric values to the status quo.
        """
        self.parameter_name = parameter_name
        self.metric_name = metric_name
        self._display_sampled = display_sampled
        self.relativize = relativize

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> PlotlyAnalysisCard:
        if experiment is None:
            raise UserInputError("SlicePlot requires an Experiment")

        relevant_adapter = extract_relevant_adapter(
            experiment=experiment,
            generation_strategy=generation_strategy,
            adapter=adapter,
        )

        metric_name = self.metric_name or select_metric(experiment=experiment)

        df = _prepare_data(
            experiment=experiment,
            model=relevant_adapter,
            parameter_name=self.parameter_name,
            metric_name=metric_name,
            relativize=self.relativize,
        )

        fig = _prepare_plot(
            df=df,
            parameter_name=self.parameter_name,
            metric_name=metric_name,
            log_x=is_axis_log_scale(
                parameter=experiment.search_space.parameters[self.parameter_name]
            ),
            display_sampled=self._display_sampled,
            is_relative=self.relativize,
        )

        return create_plotly_analysis_card(
            name=self.__class__.__name__,
            title=f"{self.parameter_name} vs. {metric_name}",
            subtitle=(
                "The slice plot provides a one-dimensional view of predicted "
                f"outcomes for {metric_name} as a function of a single parameter, "
                "while keeping all other parameters fixed at their status_quo "
                "value (or mean value if status_quo is unavailable). "
                "This visualization helps in understanding the sensitivity and "
                "impact of changes in the selected parameter on the predicted "
                "metric outcomes."
            ),
            df=df,
            fig=fig,
        )


def compute_slice_adhoc(
    parameter_name: str,
    experiment: Experiment,
    generation_strategy: GenerationStrategy | None = None,
    adapter: Adapter | None = None,
    metric_name: str | None = None,
    display_sampled: bool = True,
    relativize: bool = False,
) -> AnalysisCardBase:
    """
    Helper method to expose adhoc slice plotting. Only for advanced users in
    a notebook setting.

    Args:
        parameter_name: The name of the parameter to plot on the x-axis.
        experiment: The experiment to source data from.
        generation_strategy: Optional. The generation strategy to extract the adapter
            from.
        adapter: Optional. The adapter to use for predictions.
        metric_name: The name of the metric to plot on the y-axis. If not specified
            the objective will be used.
        display_sampled: If True, plot "x"s at x coordinates which have been sampled
            in at least one trial.
        relativize: If True, relativize the metric values to the status quo.
    """

    analysis = SlicePlot(
        parameter_name=parameter_name,
        metric_name=metric_name,
        display_sampled=display_sampled,
        relativize=relativize,
    )

    return analysis.compute(
        experiment=experiment,
        generation_strategy=generation_strategy,
        adapter=adapter,
    )


def _prepare_data(
    experiment: Experiment,
    model: Adapter,
    parameter_name: str,
    metric_name: str,
    relativize: bool,
) -> pd.DataFrame:
    sampled_xs = [
        {
            "parameter_value": arm.parameters[parameter_name],
            "arm_name": arm.name,
            "trial_index": trial.index,
        }
        for trial in experiment.trials.values()
        if trial.status in STATUSES_EXPECTING_DATA  # running, completed, early stopped
        for arm in trial.arms
        # Exclude parameter values which are not valid (ex. None when the parameter is
        # not known in a status quo arm).
        if experiment.search_space.parameters[parameter_name].validate(
            arm.parameters[parameter_name]
        )
    ]
    # Choose which parameter values to predict points for.
    unsampled_xs = get_parameter_values(
        parameter=experiment.search_space.parameters[parameter_name]
    )
    xs = [*[sample["parameter_value"] for sample in sampled_xs], *unsampled_xs]

    # Construct observation features for each parameter value previously chosen by
    # fixing all other parameters to their status-quo value or mean.
    features = [
        ObservationFeatures(
            parameters={
                parameter_name: x,
                **{
                    parameter.name: select_fixed_value(parameter=parameter)
                    for parameter in experiment.search_space.parameters.values()
                    if parameter.name != parameter_name
                },
            }
        )
        for x in xs
    ]

    predictions = model.predict(observation_features=features)

    df = none_throws(
        pd.DataFrame.from_records(
            [
                {
                    parameter_name: xs[i],
                    f"{metric_name}_mean": predictions[0][metric_name][i],
                    f"{metric_name}_sem": predictions[1][metric_name][metric_name][i]
                    ** 0.5,  # Convert the variance to the SEM
                    "sampled": xs[i]
                    in [sample["parameter_value"] for sample in sampled_xs],
                    "arm_name": sampled_xs[i]["arm_name"]
                    if i < len(sampled_xs)
                    else "unsampled",
                    "trial_index": sampled_xs[i]["trial_index"]
                    if i < len(sampled_xs)
                    else -1,
                }
                for i in range(len(xs))
            ]
        ).drop_duplicates()
    ).sort_values(by=parameter_name)

    if relativize:
        target_trial_index = none_throws(get_target_trial_index(experiment=experiment))
        df = relativize_data(
            experiment=experiment,
            df=df,
            metric_names=[metric_name],
            is_raw_data=False,
            trial_index=None,
            trial_statuses=None,
            target_trial_index=target_trial_index,
        )

    return df


def _prepare_plot(
    df: pd.DataFrame,
    parameter_name: str,
    metric_name: str,
    log_x: bool,
    display_sampled: bool,
    is_relative: bool = False,
) -> go.Figure:
    x = df[parameter_name].tolist()
    y = df[f"{metric_name}_mean"].tolist()

    # Convert the SEMs to 95% confidence intervals
    y_upper = (
        df[f"{metric_name}_mean"] + Z_SCORE_95_CI * df[f"{metric_name}_sem"]
    ).tolist()
    y_lower = (
        df[f"{metric_name}_mean"] - Z_SCORE_95_CI * df[f"{metric_name}_sem"]
    ).tolist()

    # Draw a line at the mean and a shaded region between the upper and lower bounds
    line = go.Scatter(
        x=x,
        y=y,
        line={"color": AX_BLUE},
        mode="lines",
        name=metric_name,
        showlegend=False,
    )
    error_band = go.Scatter(
        # Concatenate x values in reverse order to create a closed polygon
        x=x + x[::-1],
        # Concatenate upper and lower bounds in reverse order
        y=y_upper + y_lower[::-1],
        fill="toself",
        fillcolor=get_scatter_point_color(hex_color=AX_BLUE, ci_transparency=True),
        line={"color": "rgba(255,255,255,0)"},  # Make "line" transparent
        hoverinfo="skip",
        showlegend=False,
    )

    fig = go.Figure(
        [line, error_band],
        layout=go.Layout(
            xaxis_title=truncate_label(label=parameter_name),
            yaxis_title=truncate_label(label=metric_name),
            yaxis={"tickformat": ".2%"} if is_relative else None,
        ),
    )

    if display_sampled:
        sampled = df[df["sampled"]]
        x_sampled = sampled[parameter_name].tolist()
        y_sampled = sampled[f"{metric_name}_mean"].tolist()

        samples = go.Scatter(
            x=x_sampled,
            y=y_sampled,
            mode="markers",
            marker={
                "symbol": "x",
                "color": "black",
            },
            name=f"Sampled {parameter_name}",
            showlegend=False,
        )

        fig.add_trace(samples)

    # Set the x-axis scale to log if relevant
    if log_x:
        fig.update_xaxes(
            type="log",
            range=[
                math.log10(df[parameter_name].min()),
                math.log10(df[parameter_name].max()),
            ],
        )
    else:
        fig.update_xaxes(range=[df[parameter_name].min(), df[parameter_name].max()])

    return fig



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable, Mapping, Sequence
from copy import deepcopy
from logging import Logger
from typing import Any

import numpy as np
import numpy.typing as npt
import torch
from ax.adapter.adapter_utils import (
    array_to_observation_data,
    extract_objective_thresholds,
    extract_objective_weights,
    extract_outcome_constraints,
    extract_parameter_constraints,
    extract_risk_measure,
    extract_search_space_digest,
    get_fixed_features,
    observation_data_to_array,
    observation_features_to_array,
    parse_observation_features,
    pending_observations_as_array_list,
    prep_pairwise_data,
    process_contextual_datasets,
    SearchSpaceDigest,
    transform_callback,
    validate_and_apply_final_transform,
)
from ax.adapter.base import Adapter, DataLoaderConfig, gen_arms, GenResults
from ax.adapter.data_utils import ExperimentData, extract_experiment_data
from ax.adapter.transforms.base import Transform
from ax.adapter.transforms.cast import Cast
from ax.core.arm import Arm
from ax.core.auxiliary import AuxiliaryExperimentPurpose
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generator_run import extract_arm_predictions
from ax.core.metric import Metric
from ax.core.observation import (
    Observation,
    ObservationData,
    ObservationFeatures,
    separate_observations,
)
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
    PreferenceOptimizationConfig,
)
from ax.core.outcome_constraint import (
    ComparisonOp,
    ObjectiveThreshold,
    ScalarizedOutcomeConstraint,
)
from ax.core.search_space import SearchSpace
from ax.core.types import TCandidateMetadata, TModelPredictArm
from ax.exceptions.core import DataRequiredError, UnsupportedError, UserInputError
from ax.exceptions.generation_strategy import OptimizationConfigRequired
from ax.generators.torch.botorch import LegacyBoTorchGenerator
from ax.generators.torch.botorch_modular.generator import BoTorchGenerator
from ax.generators.torch.botorch_moo import MultiObjectiveLegacyBoTorchGenerator
from ax.generators.torch.botorch_moo_defaults import infer_objective_thresholds
from ax.generators.torch_base import TorchGenerator, TorchOptConfig
from ax.generators.types import TConfig
from ax.utils.common.constants import Keys
from ax.utils.common.logger import get_logger
from botorch.utils.datasets import MultiTaskDataset, SupervisedDataset
from pyre_extensions import none_throws
from torch import Tensor

logger: Logger = get_logger(__name__)

FIT_MODEL_ERROR = "Generator must be fit before {action}."


class TorchAdapter(Adapter):
    """An adapter for using torch-based generators.

    Specifies an interface that is implemented by TorchGenerator. In particular,
    generator should have methods fit, predict, and gen. See TorchGenerator for the
    API for each of these methods.

    Requires that all parameters have been transformed to RangeParameters
    or FixedParameters with float type and no log scale.

    This class converts Ax parameter types to torch tensors before passing
    them to the generator.
    """

    def __init__(
        self,
        *,
        experiment: Experiment,
        generator: TorchGenerator,
        search_space: SearchSpace | None = None,
        data: Data | None = None,
        transforms: Sequence[type[Transform]] | None = None,
        transform_configs: Mapping[str, TConfig] | None = None,
        optimization_config: OptimizationConfig | None = None,
        expand_model_space: bool = True,
        fit_tracking_metrics: bool = True,
        fit_on_init: bool = True,
        default_model_gen_options: TConfig | None = None,
        torch_device: torch.device | None = None,
        data_loader_config: DataLoaderConfig | None = None,
        fit_out_of_design: bool | None = None,
        fit_abandoned: bool | None = None,
        fit_only_completed_map_metrics: bool | None = None,
    ) -> None:
        """In addition to common arguments documented in the base ``Adapter`` class,
        ``TorchAdapter`` accepts the following arguments.

        Args:
            default_model_gen_options: A dictionary of default options to use
                during candidate generation. These will be overridden by any
                `model_gen_options` passed to the `Adapter.gen` method.
            torch_device: The device to use for any torch tensors and operations
                on these tensors.
            data_loader_config: A DataLoaderConfig of options for loading data. See the
                docstring of DataLoaderConfig for more details.
            fit_out_of_design: Overwrites `data_loader_config.fit_out_of_design` if
                not None.
            fit_abandoned: Overwrites `data_loader_config.fit_abandoned` if not None.
            fit_only_completed_map_metrics: If not None, overwrites
                `data_loader_config.fit_only_completed_map_metrics`.
        """
        self.device: torch.device | None = torch_device
        self._default_model_gen_options: TConfig = default_model_gen_options or {}

        self.is_moo_problem: bool = False
        self.preference_profile_name: str | None = None
        if optimization_config or (experiment and experiment.optimization_config):
            optimization_config = none_throws(
                optimization_config or experiment.optimization_config
            )
            # Handle init for multi-objective optimization.
            self.is_moo_problem = optimization_config.is_moo_problem

            if isinstance(optimization_config, PreferenceOptimizationConfig):
                self.preference_profile_name = (
                    optimization_config.preference_profile_name
                )

        # Tracks last experiment data used to fit the generator, to skip
        # generator fitting when it's not necessary.
        self._last_experiment_data: ExperimentData | None = None

        # These are set in _fit.
        self.parameters: list[str] = []
        self.outcomes: list[str] = []

        super().__init__(
            experiment=experiment,
            search_space=search_space,
            data=data,
            generator=generator,
            transforms=transforms,
            transform_configs=transform_configs,
            optimization_config=optimization_config,
            expand_model_space=expand_model_space,
            fit_tracking_metrics=fit_tracking_metrics,
            fit_on_init=fit_on_init,
            data_loader_config=data_loader_config,
            fit_out_of_design=fit_out_of_design,
            fit_abandoned=fit_abandoned,
            fit_only_completed_map_metrics=fit_only_completed_map_metrics,
        )

        # Re-assign self.generator for more precise typing.
        self.generator: TorchGenerator = generator

    def feature_importances(self, metric_name: str) -> dict[str, float]:
        importances_tensor = self.generator.feature_importances()
        importances_dict = dict(zip(self.outcomes, importances_tensor, strict=True))
        importances_arr = importances_dict[metric_name].flatten()
        return dict(zip(self.parameters, importances_arr, strict=True))

    def infer_objective_thresholds(
        self,
        search_space: SearchSpace | None = None,
        optimization_config: OptimizationConfig | None = None,
        fixed_features: ObservationFeatures | None = None,
    ) -> list[ObjectiveThreshold]:
        """Infer objective thresholds.

        This method is only applicable for Multi-Objective optimization problems.

        This method uses the generator-estimated Pareto frontier over the in-sample
        points to infer absolute (not relativized) objective thresholds.

        This uses a heuristic that sets the objective threshold to be a scaled nadir
        point, where the nadir point is scaled back based on the range of each
        objective across the current in-sample Pareto frontier.
        """
        if not self.is_moo_problem:
            raise UnsupportedError(
                "Objective thresholds are only supported for multi-objective "
                "optimization."
            )

        search_space = (search_space or self._search_space).clone()
        base_gen_args = self._get_transformed_gen_args(
            search_space=search_space,
            optimization_config=optimization_config,
            fixed_features=fixed_features,
        )
        # Get transformed args from TorchAdapter.
        search_space_digest, torch_opt_config = self._get_transformed_model_gen_args(
            search_space=base_gen_args.search_space,
            fixed_features=base_gen_args.fixed_features,
            pending_observations={},
            optimization_config=base_gen_args.optimization_config,
        )
        if torch_opt_config.risk_measure is not None:
            raise UnsupportedError(
                "`infer_objective_thresholds` does not support risk measures."
            )
        # Infer objective thresholds.
        if isinstance(self.generator, MultiObjectiveLegacyBoTorchGenerator):
            model = self.generator.model
            Xs = self.generator.Xs
        elif isinstance(self.generator, BoTorchGenerator):
            model = self.generator.surrogate.model
            Xs = self.generator.surrogate.Xs
        else:
            raise UnsupportedError(
                "Generator must be a MultiObjectiveLegacyBoTorchGenerator or an "
                "appropriate Modular Botorch Generator to infer_objective_thresholds. "
                f"Found {type(self.generator)}."
            )

        obj_thresholds = infer_objective_thresholds(
            model=none_throws(model),
            objective_weights=torch_opt_config.objective_weights,
            bounds=search_space_digest.bounds,
            outcome_constraints=torch_opt_config.outcome_constraints,
            linear_constraints=torch_opt_config.linear_constraints,
            fixed_features=torch_opt_config.fixed_features,
            Xs=Xs,
        )

        return self._untransform_objective_thresholds(
            objective_thresholds=obj_thresholds,
            objective_weights=torch_opt_config.objective_weights,
            opt_config_metrics=torch_opt_config.opt_config_metrics,
            fixed_features=torch_opt_config.fixed_features,
        )

    def model_best_point(
        self,
        search_space: SearchSpace | None = None,
        optimization_config: OptimizationConfig | None = None,
        pending_observations: dict[str, list[ObservationFeatures]] | None = None,
        fixed_features: ObservationFeatures | None = None,
        model_gen_options: TConfig | None = None,
    ) -> tuple[Arm, TModelPredictArm | None] | None:
        # Get modifiable versions
        if search_space is None:
            search_space = self._search_space
        search_space = search_space.clone()

        base_gen_args = self._get_transformed_gen_args(
            search_space=search_space,
            optimization_config=optimization_config,
            pending_observations=pending_observations,
            fixed_features=fixed_features,
        )
        search_space_digest, torch_opt_config = self._get_transformed_model_gen_args(
            search_space=base_gen_args.search_space,
            pending_observations=base_gen_args.pending_observations,
            fixed_features=base_gen_args.fixed_features,
            model_gen_options=model_gen_options,
            optimization_config=base_gen_args.optimization_config,
        )
        try:
            xbest = self.generator.best_point(
                search_space_digest=search_space_digest,
                torch_opt_config=torch_opt_config,
            )
        except NotImplementedError:
            xbest = None

        if xbest is None:
            return None

        best_obsf = ObservationFeatures(
            parameters={p: float(xbest[i]) for i, p in enumerate(self.parameters)}
        )

        for t in reversed(list(self.transforms.values())):
            best_obsf = t.untransform_observation_features([best_obsf])[0]

        best_point_predictions = extract_arm_predictions(
            model_predictions=self.predict([best_obsf]), arm_idx=0
        )

        best_arms, _ = gen_arms(
            observation_features=[best_obsf],
            arms_by_signature=self._arms_by_signature,
        )
        best_arm = best_arms[0]

        return best_arm, best_point_predictions

    def _array_callable_to_tensor_callable(
        self,
        array_func: Callable[[npt.NDArray], npt.NDArray],
    ) -> Callable[[Tensor], Tensor]:
        tensor_func: Callable[[Tensor], Tensor] = lambda x: (
            self._array_to_tensor(array_func(x.detach().cpu().clone().numpy()))
        )
        return tensor_func

    def _array_to_tensor(self, array: npt.NDArray | list[float]) -> Tensor:
        return torch.as_tensor(array, dtype=torch.double, device=self.device)

    def _convert_experiment_data(
        self,
        experiment_data: ExperimentData,
        outcomes: list[str],
        parameters: list[str],
        search_space_digest: SearchSpaceDigest | None,
    ) -> tuple[
        list[SupervisedDataset], list[str], list[list[TCandidateMetadata]] | None
    ]:
        """Converts ``ExperimentData`` to a dictionary of ``Dataset`` containers, a list
        of outcomes -- in the same order as the datasets -- and candidate metadata.
        The rows that have missing / NaN mean observations are dropped before
        constructing the dataset for the corresponding outcome.

        Args:
            experiment_data: A container of two dataframes ``arm_data`` and
                ``observation_data``, containing parameterizations, observations,
                and metadata extracted from ``Trial``s and ``Data`` of the experiment.
            outcomes: The names of the outcomes to extract observations for.
            parameters: The names of the parameters to extract. Any additional columns
                of ``arm_data`` that is not included in `parameters` will be ignored.
            search_space_digest: Optional ``SearchSpaceDigest`` containing information
                about the search space. This is used to convert datasets into a
                ``MultiTaskDataset`` where applicable.

        Returns:
            - A list of ``Dataset`` objects. The datasets will be for the set of
                outcomes specified in ``outcomes``, not necessarily in that order.
                Some outcomes will be grouped into a single dataset if there are
                contextual datasets.
            - A list of outcomes in the order that they appear in the datasets,
                accounting for reordering made necessary by contextual datasets.
            - An optional list of lists of candidate metadata. Each inner list
                corresponds to one outcome. Each element in the inner list corresponds
                to one observation.
                NOTE: Candidate metadata is currently only utilized in TRBO generator.
        """
        if len(outcomes) == 0:
            return [], [], None
        arm_data = experiment_data.arm_data
        obs_data = experiment_data.observation_data
        obs_data_mean = obs_data["mean"]
        sems_df = obs_data["sem"]
        # Check for duplication between parameter names and metric names
        obs_mean_cols = set(obs_data_mean.columns)
        arm_data_cols = set(arm_data.columns)
        duplicated_names = obs_mean_cols.intersection(arm_data_cols)

        # Join mean & arms to align and repeat the arm rows if necessary.
        # Add suffix if there are duplicate column names.
        mean_and_params = obs_data_mean.join(
            arm_data, how="left", lsuffix="_metric", rsuffix="_parameter"
        )
        # Reindex to only trial_index and arm_name, to move
        # any progression columns out of the index.
        levels_to_move = list(
            set(mean_and_params.index.names).difference({"trial_index", "arm_name"})
        )
        if len(levels_to_move) > 0:
            # This is a copy of the original df. We can modify in-place for cheaper.
            mean_and_params.reset_index(level=levels_to_move, inplace=True)
        # This will include the progression if it is in parameters.
        # This is also tolerant to missing columns, which is relevant for TL.
        params_np = mean_and_params.filter(
            [i + "_parameter" if i in duplicated_names else i for i in parameters]
        ).to_numpy()
        trial_indices_np = mean_and_params.index.get_level_values(
            "trial_index"
        ).to_numpy()
        metadata = mean_and_params["metadata"]
        datasets: list[SupervisedDataset] = []
        candidate_metadata = []
        for outcome in outcomes:
            outcome_col_name = (
                outcome + "_metric" if outcome in duplicated_names else outcome
            )
            if outcome_col_name not in mean_and_params:
                raise DataRequiredError(
                    f"Attempting to extract a dataset for {outcome=} but no "
                    "corresponding data was found in the experiment data."
                )
            # Drop NaN columns from means & corresponding params.
            outcome_means = mean_and_params[outcome_col_name].to_numpy()
            to_keep = ~np.isnan(outcome_means)
            Y = torch.from_numpy(outcome_means[to_keep]).double().view(-1, 1)
            X = torch.from_numpy(params_np[to_keep]).double()
            sem = sems_df[outcome].to_numpy()[to_keep]
            if np.all(np.isnan(sem)):
                Yvar = None
            else:
                Yvar = torch.from_numpy(sem).double().square().view(-1, 1)
            group_indices = torch.from_numpy(trial_indices_np[to_keep])
            if outcome == Keys.PAIRWISE_PREFERENCE_QUERY.value:
                dataset = prep_pairwise_data(
                    X=X,
                    Y=Y.to(dtype=torch.long),
                    group_indices=group_indices,
                    outcome=outcome,
                    parameters=parameters,
                )
            else:
                dataset = SupervisedDataset(
                    X=X,
                    Y=Y,
                    Yvar=Yvar,
                    feature_names=parameters,
                    outcome_names=[outcome],
                    group_indices=group_indices,
                )
            datasets.append(dataset)
            candidate_metadata.append(metadata.loc[to_keep].to_list())

        # If the search space digest specifies a task feature,
        # convert the datasets into MultiTaskDataset.
        if search_space_digest is not None and (
            task_features := search_space_digest.task_features
        ):
            if len(task_features) > 1:
                raise UnsupportedError("Multiple task features are not supported.")
            target_task_value = search_space_digest.target_values[task_features[0]]
            # Both the MTGP and the MTDataset expect integer valued task features.
            # Check that they're indeed integers.
            task_choices = search_space_digest.discrete_choices[task_features[0]]
            if any(int(t) != t for t in task_choices):
                raise ValueError(
                    "The values of the task feature must be integers. "
                    "This is often accomplished using a "
                    "TaskChoiceToIntTaskChoice transform. "
                    "Check that the generator is using the correct set of transforms. "
                    f"Got {task_choices=}."
                )
            datasets = [
                MultiTaskDataset.from_joint_dataset(
                    dataset=dataset,
                    task_feature_index=task_features[0],
                    target_task_value=int(target_task_value),
                )
                for dataset in datasets
            ]
        # Check if there is a `parameter_decomposition` experiment property to
        # decide whether it is a contextual experiment.
        if self._experiment_properties.get("parameter_decomposition", None) is not None:
            # Convert to a list of ContextualDateset for contextual experiments.
            # pyre-ignore [9]: ContextualDataset is a subclass of SupervisedDataset.
            datasets = process_contextual_datasets(
                datasets=datasets,
                outcomes=outcomes,
                parameter_decomposition=self._experiment_properties[
                    "parameter_decomposition"
                ],
                metric_decomposition=self._experiment_properties.get(
                    "metric_decomposition", None
                ),
            )

        # Get the order of outcomes
        ordered_outcomes = []
        for d in datasets:
            ordered_outcomes.extend(d.outcome_names)
        # Re-order candidate metadata
        if not metadata.isnull().all():
            ordered_metadata = []
            for outcome in ordered_outcomes:
                ordered_metadata.append(candidate_metadata[outcomes.index(outcome)])
        else:
            ordered_metadata = None

        return datasets, ordered_outcomes, ordered_metadata

    def _cross_validate(
        self,
        search_space: SearchSpace,
        cv_training_data: ExperimentData,
        cv_test_points: list[ObservationFeatures],
        use_posterior_predictive: bool = False,
    ) -> list[ObservationData]:
        """Make predictions at ``cv_test_points`` using only the data
        in ``cv_training_data``.
        """
        if self.parameters is None:
            raise ValueError(FIT_MODEL_ERROR.format(action="_cross_validate"))
        datasets, _, search_space_digest = self._get_fit_args(
            search_space=search_space,
            experiment_data=cv_training_data,
            update_outcomes_and_parameters=False,
        )
        X_test = torch.tensor(
            [[obsf.parameters[p] for p in self.parameters] for obsf in cv_test_points],
            dtype=torch.double,
            device=self.device,
        )
        # Use the generator to do the cross validation
        f_test, cov_test = self.generator.cross_validate(
            datasets=datasets,
            X_test=torch.as_tensor(X_test, dtype=torch.double, device=self.device),
            search_space_digest=search_space_digest,
            use_posterior_predictive=use_posterior_predictive,
        )
        # Convert array back to ObservationData
        return array_to_observation_data(
            f=f_test.detach().cpu().clone().numpy(),
            cov=cov_test.detach().cpu().clone().numpy(),
            outcomes=self.outcomes,
        )

    def evaluate_acquisition_function(
        self,
        observation_features: (
            list[ObservationFeatures] | list[list[ObservationFeatures]]
        ),
        search_space: SearchSpace | None = None,
        optimization_config: OptimizationConfig | None = None,
        pending_observations: dict[str, list[ObservationFeatures]] | None = None,
        fixed_features: ObservationFeatures | None = None,
        acq_options: dict[str, Any] | None = None,
    ) -> list[float]:
        """Evaluate the acquisition function for given set of observation
        features.

        Args:
            observation_features: Either a list or a list of lists of observation
                features, representing parameterizations, for which to evaluate the
                acquisition function. If a single list is passed, the acquisition
                function is evaluated for each observation feature. If a list of lists
                is passed each element (itself a list of observation features)
                represents a batch of points for which to evaluate the joint acquisition
                value.
            search_space: Search space for fitting the generator.
            optimization_config: Optimization config defining how to optimize
                the generator.
            pending_observations: A map from metric name to pending observations for
                that metric.
            fixed_features: An ObservationFeatures object containing any features that
                should be fixed at specified values during generation.
            acq_options: Keyword arguments used to construct the acquisition function.

        Returns:
            A list of acquisition function values, in the same order as the
            input observation features.
        """
        search_space = search_space or self._search_space
        optimization_config = optimization_config or self._optimization_config
        if optimization_config is None:
            raise ValueError(
                "The `optimization_config` must be specified either while initializing "
                "the Adapter or to the `evaluate_acquisition_function` call."
            )
        # pyre-ignore Incompatible parameter type [9]
        obs_feats: list[list[ObservationFeatures]] = deepcopy(observation_features)
        if not isinstance(obs_feats[0], list):
            obs_feats = [[obs] for obs in obs_feats]

        for t in self.transforms.values():
            for i, batch in enumerate(obs_feats):
                obs_feats[i] = t.transform_observation_features(batch)

        base_gen_args = self._get_transformed_gen_args(
            search_space=search_space,
            optimization_config=optimization_config,
            pending_observations=pending_observations,
            fixed_features=fixed_features,
        )

        return self._evaluate_acquisition_function(
            observation_features=obs_feats,
            search_space=base_gen_args.search_space,
            optimization_config=none_throws(base_gen_args.optimization_config),
            pending_observations=base_gen_args.pending_observations,
            fixed_features=base_gen_args.fixed_features,
            acq_options=acq_options,
        )

    def _evaluate_acquisition_function(
        self,
        observation_features: list[list[ObservationFeatures]],
        search_space: SearchSpace,
        optimization_config: OptimizationConfig,
        pending_observations: dict[str, list[ObservationFeatures]] | None = None,
        fixed_features: ObservationFeatures | None = None,
        acq_options: dict[str, Any] | None = None,
    ) -> list[float]:
        if not self.parameters:
            raise RuntimeError(
                FIT_MODEL_ERROR.format(action="_evaluate_acquisition_function")
            )
        search_space_digest, torch_opt_config = self._get_transformed_model_gen_args(
            search_space=search_space,
            pending_observations=pending_observations or {},
            fixed_features=fixed_features or ObservationFeatures({}),
            optimization_config=optimization_config,
        )
        X = np.array(
            [
                observation_features_to_array(self.parameters, obsf)
                for obsf in observation_features
            ]
        )
        evals = self.generator.evaluate_acquisition_function(
            X=self._array_to_tensor(X),
            search_space_digest=search_space_digest,
            torch_opt_config=torch_opt_config,
            acq_options=acq_options,
        )
        return evals.tolist()

    def _update_w_aux_exp_datasets(
        self, datasets: list[SupervisedDataset]
    ) -> list[SupervisedDataset]:
        aux_datasets = []
        # Extract datasets needed from auxiliary experiments
        # For preference exploration
        optimization_config = self._optimization_config
        if isinstance(optimization_config, PreferenceOptimizationConfig):
            target_pe_aux_exp_name = optimization_config.preference_profile_name
            target_pe_aux_exp = none_throws(
                self._experiment.find_auxiliary_experiment_by_name(
                    purpose=AuxiliaryExperimentPurpose.PE_EXPERIMENT,
                    auxiliary_experiment_name=target_pe_aux_exp_name,
                    raise_if_not_found=True,
                )
            )
            # This is the name of the PE experiment on which we fit
            # the learned objective preference model
            self.preference_profile_name = target_pe_aux_exp_name
            pe_exp, pe_data = target_pe_aux_exp.experiment, target_pe_aux_exp.data

            if pe_data.df.empty:
                raise DataRequiredError(
                    "No data found in the auxiliary preference exploration "
                    "experiment. Play the preference game first or use another"
                    "preference profile with recorded preference data."
                )

            pe_experiment_data = extract_experiment_data(
                experiment=pe_exp,
                data=pe_data,
                data_loader_config=self._data_loader_config,
            )
            pe_exp_param_names = list(pe_exp.search_space.parameters.keys())

            # Get all relevant information on the parameters
            pe_ssd = extract_search_space_digest(
                search_space=pe_exp.search_space, param_names=pe_exp_param_names
            )
            pe_outcomes = [Keys.PAIRWISE_PREFERENCE_QUERY.value]

            pe_datasets, _, _ = self._convert_experiment_data(
                experiment_data=pe_experiment_data,
                outcomes=pe_outcomes,
                parameters=pe_exp_param_names,
                search_space_digest=pe_ssd,
            )
            aux_datasets.extend(pe_datasets)
        return datasets + aux_datasets

    def _get_fit_args(
        self,
        search_space: SearchSpace,
        experiment_data: ExperimentData,
        update_outcomes_and_parameters: bool,
    ) -> tuple[
        list[SupervisedDataset],
        list[list[TCandidateMetadata]] | None,
        SearchSpaceDigest,
    ]:
        """Helper for consolidating some common argument processing between
        ``fit`` and ``cross_validate`` methods. Extracts datasets and candidate metadata
        from ``experiment_data``, and ``search_space_digest`` from the ``search_space``.

        Args:
            search_space: A transformed search space for fitting the generator.
            experiment_data: A container of two dataframes ``arm_data`` and
                ``observation_data``, containing parameterizations, observations,
                and metadata extracted from ``Trial``s and ``Data`` of the experiment.
            update_outcomes_and_parameters: Whether to update `self.outcomes` with
                all outcomes found in the observations and `self.parameters` with
                all parameters in the search space. Typically only used in `_fit`.

        Returns:
            The datasets & metadata, extracted from the ``experiment_data``, and the
            ``search_space_digest``.
        """
        self._last_experiment_data = experiment_data
        if update_outcomes_and_parameters:
            self.parameters = list(search_space.parameters.keys())
            # Make sure that task feature is the last parameter. This is important
            # for heterogeneous search spaces
            if Keys.TASK_FEATURE_NAME.value in self.parameters:
                idx = self.parameters.index(Keys.TASK_FEATURE_NAME.value)
                if idx != len(self.parameters) - 1:
                    self.parameters = (
                        self.parameters[:idx]
                        + self.parameters[idx + 1 :]
                        + [Keys.TASK_FEATURE_NAME.value]
                    )
        # Only update outcomes if fitting a model on tracking metrics. Otherwise,
        # we will only fit models to the outcomes that are extracted from optimization
        # config in Adapter.__init__.
        if update_outcomes_and_parameters and self._fit_tracking_metrics:
            self.outcomes = sorted(experiment_data.metric_names)
        # Get all relevant information on the parameters
        search_space_digest = extract_search_space_digest(
            search_space=search_space, param_names=self.parameters
        )
        # Convert observations to datasets
        datasets, ordered_outcomes, candidate_metadata = self._convert_experiment_data(
            experiment_data=experiment_data,
            outcomes=self.outcomes,
            parameters=self.parameters,
            search_space_digest=search_space_digest,
        )

        # Do not support handling of auxiliary datasets in legacy BoTorch generators
        if not isinstance(self.generator, LegacyBoTorchGenerator):
            datasets = self._update_w_aux_exp_datasets(datasets=datasets)

        if update_outcomes_and_parameters:
            self.outcomes = ordered_outcomes
        else:
            assert (
                ordered_outcomes == self.outcomes
            ), f"Unexpected ordering of outcomes: {ordered_outcomes} != {self.outcomes}"
        return datasets, candidate_metadata, search_space_digest

    def _fit(
        self,
        search_space: SearchSpace,
        experiment_data: ExperimentData,
        **kwargs: Any,
    ) -> None:
        if experiment_data == self._last_experiment_data:
            logger.debug(
                "The experiment data is identical to the last experiment data "
                "used to fit the generator. Skipping generator fitting."
            )
            return
        datasets, candidate_metadata, search_space_digest = self._get_fit_args(
            search_space=search_space,
            experiment_data=experiment_data,
            update_outcomes_and_parameters=True,
        )
        self.generator.fit(
            datasets=datasets,
            search_space_digest=search_space_digest,
            candidate_metadata=candidate_metadata,
            **kwargs,
        )

    def _gen(
        self,
        n: int,
        search_space: SearchSpace,
        pending_observations: dict[str, list[ObservationFeatures]],
        fixed_features: ObservationFeatures | None,
        model_gen_options: TConfig | None = None,
        optimization_config: OptimizationConfig | None = None,
    ) -> GenResults:
        """Generate new candidates according to search_space and
        optimization_config.

        The outcome constraints should be transformed to no longer be relative.
        """
        if not self.parameters:
            raise ValueError(FIT_MODEL_ERROR.format(action="_gen"))

        # Ensure the preference model we fit is the one used in optimization_config
        if isinstance(optimization_config, PreferenceOptimizationConfig) and (
            optimization_config.preference_profile_name != self.preference_profile_name
        ):
            raise UserInputError(
                "The preference profile name in the optimization config does not match "
                "the name of the preference profile used to fit the preference model. "
                f"Expected {self.preference_profile_name} but got "
                f"{optimization_config.preference_profile_name}. "
                "Consider updating `experiment.optimization_config` and refit "
                "the model before proceeding with gen."
            )

        augmented_model_gen_options = {
            **self._default_model_gen_options,
            **(model_gen_options or {}),
        }
        search_space_digest, torch_opt_config = self._get_transformed_model_gen_args(
            search_space=search_space,
            pending_observations=pending_observations,
            fixed_features=fixed_features,
            model_gen_options=augmented_model_gen_options,
            optimization_config=optimization_config,
        )

        # Generate the candidates
        gen_results = self.generator.gen(
            n=n,
            search_space_digest=search_space_digest,
            torch_opt_config=torch_opt_config,
        )

        gen_metadata = dict(gen_results.gen_metadata)
        if (
            isinstance(optimization_config, MultiObjectiveOptimizationConfig)
            and gen_metadata.get("objective_thresholds", None) is not None
        ):
            # If objective_thresholds are supplied by the user, then the transformed
            # user-specified objective thresholds are in gen_metadata. Otherwise,
            # if using a hypervolume based acquisition function, then
            # the inferred objective thresholds are in gen_metadata.
            gen_metadata["objective_thresholds"] = (
                self._untransform_objective_thresholds(
                    objective_thresholds=gen_metadata["objective_thresholds"],
                    objective_weights=torch_opt_config.objective_weights,
                    opt_config_metrics=torch_opt_config.opt_config_metrics,
                    fixed_features=torch_opt_config.fixed_features,
                )
            )

        # Transform array to observations
        observation_features = self._array_to_observation_features(
            X=gen_results.points.detach().cpu().clone().numpy(),
            candidate_metadata=gen_results.candidate_metadata,
        )
        try:
            xbest = self.generator.best_point(
                search_space_digest=search_space_digest,
                torch_opt_config=torch_opt_config,
            )
        except NotImplementedError:
            xbest = None

        best_obsf = None
        if xbest is not None:
            best_obsf = ObservationFeatures(
                parameters={
                    p: float(x) for p, x in zip(self.parameters, xbest, strict=True)
                }
            )

        return GenResults(
            observation_features=observation_features,
            weights=gen_results.weights.tolist(),
            best_observation_features=best_obsf,
            gen_metadata=gen_metadata,
        )

    def _predict(
        self,
        observation_features: list[ObservationFeatures],
        use_posterior_predictive: bool = False,
    ) -> list[ObservationData]:
        if not self.parameters:
            raise ValueError(FIT_MODEL_ERROR.format(action="_model_predict"))
        # Convert observation features to array
        X = observation_features_to_array(self.parameters, observation_features)
        f, cov = self.generator.predict(
            X=self._array_to_tensor(X),
            use_posterior_predictive=use_posterior_predictive,
        )
        f = f.detach().cpu().clone().numpy()
        cov = cov.detach().cpu().clone().numpy()
        if f.shape[-2] != X.shape[-2]:
            raise NotImplementedError(
                "Expected same number of predictions as the number of inputs but got "
                f"predictions of shape {f.shape} for inputs of shape {X.shape}. "
                "This was likely due to the use of one-to-many input transforms -- "
                "typically used for robust optimization -- which are not supported in"
                "TorchAdapter.predict."
            )
        # Convert resulting arrays to observations
        return array_to_observation_data(f=f, cov=cov, outcomes=self.outcomes)

    def _array_to_observation_features(
        self,
        X: npt.NDArray,
        candidate_metadata: Sequence[TCandidateMetadata] | None,
    ) -> list[ObservationFeatures]:
        return parse_observation_features(
            X=X, param_names=self.parameters, candidate_metadata=candidate_metadata
        )

    def _transform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> Tensor:
        """Apply terminal transform to given observation features and return result
        as an N x D array of points.
        """
        try:
            tobfs = np.array(
                [
                    [float(of.parameters[p]) for p in self.parameters]
                    for of in observation_features
                ]
            )
        except (KeyError, TypeError):
            raise ValueError("Invalid formatting of observation features.")
        return self._array_to_tensor(tobfs)

    def _get_transformed_model_gen_args(
        self,
        search_space: SearchSpace,
        pending_observations: dict[str, list[ObservationFeatures]],
        fixed_features: ObservationFeatures | None,
        model_gen_options: TConfig | None = None,
        optimization_config: OptimizationConfig | None = None,
    ) -> tuple[SearchSpaceDigest, TorchOptConfig]:
        # Validation
        if not self.parameters:
            raise ValueError(FIT_MODEL_ERROR.format(action="_gen"))
        # Extract search space info
        search_space_digest = extract_search_space_digest(
            search_space=search_space, param_names=self.parameters
        )
        if optimization_config is None:
            raise OptimizationConfigRequired(
                f"{self.__class__.__name__} requires an OptimizationConfig "
                "to be specified"
            )

        validate_transformed_optimization_config(optimization_config, self.outcomes)
        objective_weights = extract_objective_weights(
            objective=optimization_config.objective, outcomes=self.outcomes
        )
        outcome_constraints = extract_outcome_constraints(
            outcome_constraints=optimization_config.outcome_constraints,
            outcomes=self.outcomes,
        )
        linear_constraints = extract_parameter_constraints(
            search_space.parameter_constraints, self.parameters
        )
        fixed_features_dict = get_fixed_features(fixed_features, self.parameters)

        if isinstance(optimization_config, MultiObjectiveOptimizationConfig):
            objective_thresholds = extract_objective_thresholds(
                objective_thresholds=optimization_config.objective_thresholds,
                objective=optimization_config.objective,
                outcomes=self.outcomes,
            )
        else:
            objective_thresholds = None
        opt_config_metrics = optimization_config.metrics

        pending_array = pending_observations_as_array_list(
            pending_observations, self.outcomes, self.parameters
        )
        obj_w, out_c, lin_c, pend_o, obj_t = validate_and_apply_final_transform(
            objective_weights=objective_weights,
            outcome_constraints=outcome_constraints,
            linear_constraints=linear_constraints,
            pending_observations=pending_array,
            objective_thresholds=objective_thresholds,
            final_transform=self._array_to_tensor,
        )
        rounding_func = self._array_callable_to_tensor_callable(
            transform_callback(self.parameters, self.transforms)
        )
        risk_measure = (
            optimization_config.risk_measure
            if optimization_config is not None
            else None
        )
        if risk_measure is not None:
            if not self.generator._supports_robust_optimization:
                raise UnsupportedError(
                    f"{self.generator.__class__.__name__} does not support robust "
                    "optimization. Consider using modular BoTorch generator instead."
                )
            else:
                risk_measure = extract_risk_measure(risk_measure=risk_measure)

        use_learned_objective = False
        if isinstance(optimization_config, PreferenceOptimizationConfig):
            use_learned_objective = True

        torch_opt_config = TorchOptConfig(
            objective_weights=obj_w,
            outcome_constraints=out_c,
            objective_thresholds=obj_t,
            linear_constraints=lin_c,
            fixed_features=fixed_features_dict,
            pending_observations=pend_o,
            model_gen_options=model_gen_options or {},
            rounding_func=rounding_func,
            opt_config_metrics=opt_config_metrics,
            is_moo=optimization_config.is_moo_problem,
            risk_measure=risk_measure,
            fit_out_of_design=self._data_loader_config.fit_out_of_design,
            use_learned_objective=use_learned_objective,
        )
        return search_space_digest, torch_opt_config

    def _transform_observations(
        self, observations: list[Observation]
    ) -> tuple[Tensor, Tensor, Tensor]:
        """Apply terminal transform to given observation data and return result.

        Converts a set of observations to a tuple of
            - a (n x d) array of X
            - an (n x m) array of means
            - an (n x m x m) array of covariances
        """
        observation_features, observation_data = separate_observations(observations)
        try:
            mean, cov = observation_data_to_array(
                outcomes=self.outcomes, observation_data=observation_data
            )
        except (KeyError, TypeError):
            raise ValueError("Invalid formatting of observation data.")
        X = self._transform_observation_features(
            observation_features=observation_features
        )
        return X, self._array_to_tensor(mean), self._array_to_tensor(cov)

    def _untransform_objective_thresholds(
        self,
        objective_thresholds: Tensor,
        objective_weights: Tensor,
        opt_config_metrics: dict[str, Metric],
        fixed_features: dict[int, float] | None,
    ) -> list[ObjectiveThreshold]:
        """Converts tensor-valued (possibly inferred) objective thresholds to
        ``ObjectiveThreshold`` objects, and untransforms to ensure they are
        on the same raw scale as the original optimization config.

        Args:
            objective_thresholds: A tensor of (possibly inferred) objective thresholds
                of shape `(num_metrics)`.
            objective_weights: A tensor of objective weights that denote whether each
                objective is being minimized (-1) or maximized (+1). May also include
                0 values, which represents outcome constraints and tracking metrics.
            opt_config_metrics: A dictionary mapping the metric name to the ``Metric``
                object from the original optimization config.
            fixed_features: A map {feature_index: value} for features that should be
                fixed to a particular value during generation. This typically includes
                the target trial index for multi-task applications.

        Returns:
            A list of ``ObjectiveThreshold``s on the raw, untransformed scale.
        """
        idxs = objective_weights.nonzero().view(-1).tolist()

        # Create transformed ObjectiveThresholds from tensor thresholds.
        thresholds = []
        for idx in idxs:
            sign = torch.sign(objective_weights[idx])
            thresholds.append(
                ObjectiveThreshold(
                    metric=opt_config_metrics[self.outcomes[idx]],
                    bound=float(objective_thresholds[idx].item()),
                    relative=False,
                    op=ComparisonOp.LEQ if sign < 0 else ComparisonOp.GEQ,
                )
            )
        fixed_features = fixed_features or {}
        fixed_features_obs = ObservationFeatures(
            parameters={
                name: fixed_features[i]
                for i, name in enumerate(self.parameters)
                if i in fixed_features
            }
        )

        for t in reversed(list(self.transforms.values())):
            if not isinstance(t, Cast):
                # Cast transform requires a valid hierarchical parameterization.
                # `fixed_features_obs` is incomplete, so it leads to an error.
                fixed_features_obs = t.untransform_observation_features(
                    [fixed_features_obs]
                )[0]
            thresholds = t.untransform_outcome_constraints(
                outcome_constraints=thresholds,
                fixed_features=fixed_features_obs,
            )

        return thresholds

    def _validate_observation_data(
        self, observation_data: list[ObservationData]
    ) -> None:
        if len(observation_data) == 0:
            raise ValueError(
                "Torch generators cannot be fit without observation data. Possible "
                "reasons include empty data being passed to the generator's constructor"
                " or data being excluded because it is out-of-design. Try setting "
                "`fit_out_of_design`=True during construction to fix the latter."
            )

    def _extract_observation_data(
        self,
        observation_data: list[ObservationData],
        observation_features: list[ObservationFeatures],
        parameters: list[str],
    ) -> tuple[
        dict[str, list[Tensor]],
        dict[str, list[Tensor]],
        dict[str, list[Tensor]],
        dict[str, list[TCandidateMetadata]],
        bool,
        dict[str, list[int]] | None,
    ]:
        """Extract observation features & data into tensors and metadata.

        Args:
            observation_data: A list of `ObservationData` from which to extract
                mean `Y` and variance `Yvar` observations. Must correspond 1:1 to
                the `observation_features`.
            observation_features: A list of `ObservationFeatures` from which to extract
                parameter values. Must correspond 1:1 to the `observation_data`.
            parameters: The names of the parameters to extract. Any observation features
                that are not included in `parameters` will be ignored.

        Returns:
            - A dictionary mapping metric names to lists of corresponding feature
                tensors `X`.
            - A dictionary mapping metric names to lists of corresponding mean
                observation tensors `Y`.
            - A dictionary mapping metric names to lists of corresponding variance
                observation tensors `Yvar`.
            - A dictionary mapping metric names to lists of corresponding metadata.
            - A boolean denoting whether any candidate metadata is not none.
            - A dictionary mapping metric names to lists of corresponding trial indices,
                or None if trial indices are not provided. This is used to support
                learning-curve-based modeling.
        """
        Xs: dict[str, list[Tensor]] = defaultdict(list)
        Ys: dict[str, list[Tensor]] = defaultdict(list)
        Yvars: dict[str, list[Tensor]] = defaultdict(list)
        candidate_metadata_dict: dict[str, list[TCandidateMetadata]] = defaultdict(list)
        any_candidate_metadata_is_not_none = False
        trial_indices: dict[str, list[int]] = defaultdict(list)

        at_least_one_trial_index_is_none = False
        for obsd, obsf in zip(observation_data, observation_features, strict=True):
            try:
                x = torch.tensor(
                    [obsf.parameters[p] for p in parameters],
                    dtype=torch.double,
                    device=self.device,
                )
            except (KeyError, TypeError):
                raise ValueError("Out of design points cannot be converted.")
            for metric_name, mean, var in zip(
                obsd.metric_names, obsd.means, obsd.covariance.diagonal()
            ):
                Xs[metric_name].append(x)
                Ys[metric_name].append(mean)
                Yvars[metric_name].append(var)
                if obsf.metadata is not None:
                    any_candidate_metadata_is_not_none = True
                candidate_metadata_dict[metric_name].append(obsf.metadata)
                trial_index = obsf.trial_index
                if trial_index is not None:
                    trial_indices[metric_name].append(trial_index)
                else:
                    at_least_one_trial_index_is_none = True
        if len(trial_indices) > 0 and at_least_one_trial_index_is_none:
            raise ValueError(
                "Trial indices must be provided for all observation_features "
                "or none of them."
            )

        return (
            Xs,
            Ys,
            Yvars,
            candidate_metadata_dict,
            any_candidate_metadata_is_not_none,
            trial_indices if len(trial_indices) > 0 else None,
        )


def validate_transformed_optimization_config(
    optimization_config: OptimizationConfig, outcomes: list[str]
) -> None:
    """Validate optimization config against generator fitted outcomes.

    Args:
        optimization_config: Config to validate.
        outcomes: List of metric names w/ valid generator fits.

    Raises if:
            1. In the modeling layer, absolute constraints are required, however,
               specifying relative constraints is supported. We handle this by
               either 1) transforming the observations to be relative (so that the
               constraints are absolute w.r.t relativized metrics), or 2)
               derelativizing the constraint. If relative constraints are found at
               this layer, we raise an error as likely either `Relativize` or
               `Derelativized` transforms were expected to be applied but were not.
            2. Optimization metrics are not present in generator fitted outcomes.
    """
    for c in optimization_config.outcome_constraints:
        if c.relative:
            raise ValueError(
                f"Passed {c} as a relative constraint. This likely indicates that "
                "either a `Relativize` or `Derelativize` transform was expected to be"
                " applied but was not."
            )
        if isinstance(c, ScalarizedOutcomeConstraint):
            for c_metric in c.metrics:
                if c_metric.name not in outcomes:
                    raise DataRequiredError(
                        f"Scalarized constraint metric component {c.metric.name} "
                        + "not found in fitted data."
                    )
        elif c.metric.name not in outcomes:
            raise DataRequiredError(
                f"Outcome constraint metric {c.metric.name} not found in fitted data."
            )
    obj_metric_names = [m.name for m in optimization_config.objective.metrics]
    for obj_metric_name in obj_metric_names:
        if obj_metric_name not in outcomes:
            raise DataRequiredError(
                f"Objective metric {obj_metric_name} not found in fitted data."
            )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import asyncio
import functools
import threading
import time
from collections.abc import Callable, Generator
from contextlib import contextmanager
from functools import partial
from logging import Logger
from typing import Any, Iterator, List, Type, TypeVar


MAX_WAIT_SECONDS: int = 600
T = TypeVar("T")


# pyre-fixme[3]: Return annotation cannot be `Any`.
def retry_on_exception(
    exception_types: tuple[type[Exception], ...] | None = None,
    no_retry_on_exception_types: tuple[type[Exception], ...] | None = None,
    check_message_contains: list[str] | None = None,
    retries: int = 3,
    suppress_all_errors: bool = False,
    logger: Logger | None = None,
    # pyre-fixme[2]: Parameter annotation cannot be `Any`.
    default_return_on_suppression: Any | None = None,
    wrap_error_message_in: str | None = None,
    initial_wait_seconds: int | None = None,
) -> Any | None:
    """
    A decorator for instance methods or standalone functions that makes them
    retry on failure and allows to specify on which types of exceptions the
    function should and should not retry.

    NOTE: If the argument `suppress_all_errors` is supplied and set to True,
    the error will be suppressed  and default value returned.

    Args:
        exception_types: A tuple of exception(s) types to catch in the decorated
            function. If none is provided, baseclass Exception will be used.

        no_retry_on_exception_types: Exception types to consider non-retryable even
            if their supertype appears in `exception_types` or the only exceptions to
            not retry on if no `exception_types` are specified.

        check_message_contains: A list of strings, against which to match error
            messages. If the error message contains any one of these strings,
            the exception will cause a retry. NOTE: This argument works in
            addition to `exception_types`; if those are specified, only the
            specified types of exceptions will be caught and retried on if they
            contain the strings provided as `check_message_contains`.

        retries: Number of retries to perform.

        suppress_all_errors: If true, after all the retries are exhausted, the
            error will still be suppressed and `default_return_on_suppresion`
            will be returned from the function. NOTE: If using this argument,
            the decorated function may not actually get fully executed, if
            it consistently raises an exception.

        logger: A handle for the logger to be used.

        default_return_on_suppression: If the error is suppressed after all the
            retries, then this default value will be returned from the function.
            Defaults to None.

        wrap_error_message_in: If raising the error message after all the retries,
            a string wrapper for the error message (useful for making error
            messages more user-friendly). NOTE: Format of resulting error will be:
            "<wrap_error_message_in>: <original_error_type>: <original_error_msg>",
            with the stack trace of the original message.

        initial_wait_seconds: Initial length of time to wait between failures,
            doubled after each failure up to a maximum of 10 minutes. If unspecified
            then there is no wait between retries.
    """

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def func_wrapper(func):
        # Depending on whether `func` is async or not, we use a slightly different
        # wrapper; if wrapping an async function, decorator will await it.
        # `async_actual_wrapper` and `actual_wrapper` are almost exactly the same,
        # except that the former is async and awaits the wrapped function.
        if asyncio.iscoroutinefunction(func):

            @functools.wraps(func)
            # pyre-fixme[53]: Captured variable `func` is not annotated.
            # pyre-fixme[3]: Return type must be annotated.
            # pyre-fixme[2]: Parameter must be annotated.
            async def async_actual_wrapper(*args, **kwargs):
                (
                    retry_exceptions,
                    no_retry_exceptions,
                    suppress_errors,
                ) = _validate_and_fill_defaults(
                    retry_on_exception_types=exception_types,
                    no_retry_on_exception_types=no_retry_on_exception_types,
                    suppress_errors=suppress_all_errors,
                    **kwargs,
                )

                for i in range(retries):
                    with handle_exceptions_in_retries(
                        no_retry_exceptions=no_retry_exceptions,
                        retry_exceptions=retry_exceptions,
                        suppress_errors=suppress_errors,
                        # pyre-fixme[6]: For 4th param expected `Optional[str]` but
                        #  got `Optional[List[str]]`.
                        check_message_contains=check_message_contains,
                        last_retry=i >= retries - 1,
                        logger=logger,
                        wrap_error_message_in=wrap_error_message_in,
                    ):
                        if i > 0 and initial_wait_seconds is not None:
                            wait_interval = min(
                                MAX_WAIT_SECONDS, initial_wait_seconds * 2 ** (i - 1)
                            )
                            # pyre-fixme[1001]: `asyncio.sleep(wait_interval)` is
                            #  never awaited.
                            asyncio.sleep(wait_interval)
                        return await func(*args, **kwargs)
                # If we are here, it means the retries were finished but
                # The error was suppressed. Hence return the default value provided.
                return default_return_on_suppression

            return async_actual_wrapper

        @functools.wraps(func)
        # pyre-fixme[53]: Captured variable `func` is not annotated.
        # pyre-fixme[3]: Return type must be annotated.
        # pyre-fixme[2]: Parameter must be annotated.
        def actual_wrapper(*args, **kwargs):
            (
                retry_exceptions,
                no_retry_exceptions,
                suppress_errors,
            ) = _validate_and_fill_defaults(
                retry_on_exception_types=exception_types,
                no_retry_on_exception_types=no_retry_on_exception_types,
                suppress_errors=suppress_all_errors,
                **kwargs,
            )

            for i in range(retries):
                with handle_exceptions_in_retries(
                    no_retry_exceptions=no_retry_exceptions,
                    retry_exceptions=retry_exceptions,
                    suppress_errors=suppress_errors,
                    # pyre-fixme[6]: For 4th param expected `Optional[str]` but got
                    #  `Optional[List[str]]`.
                    check_message_contains=check_message_contains,
                    last_retry=i >= retries - 1,
                    logger=logger,
                    wrap_error_message_in=wrap_error_message_in,
                ):
                    if i > 0 and initial_wait_seconds is not None:
                        wait_interval = min(
                            MAX_WAIT_SECONDS, initial_wait_seconds * 2 ** (i - 1)
                        )
                        time.sleep(wait_interval)
                    return func(*args, **kwargs)

            # If we are here, it means the retries were finished but
            # The error was suppressed. Hence return the default value provided.
            return default_return_on_suppression

        return actual_wrapper

    return func_wrapper


@contextmanager
def handle_exceptions_in_retries(
    no_retry_exceptions: tuple[type[Exception], ...],
    retry_exceptions: tuple[type[Exception], ...],
    suppress_errors: bool,
    check_message_contains: str | None,
    last_retry: bool,
    logger: Logger | None,
    wrap_error_message_in: str | None,
) -> Generator[None, None, None]:
    try:
        yield  # Perform action within the context manager.
    except no_retry_exceptions:
        raise
    except retry_exceptions as err:  # Exceptions is a tuple.
        err_msg = getattr(err, "message", repr(err))

        if not last_retry or suppress_errors:
            # We are either explicitly asked to suppress the error
            # or we have retries left.
            if logger is not None:
                # `logger.exception` automatically logs `err` and its stack trace.
                logger.exception(err)

        elif (
            not last_retry
            and check_message_contains is not None
            and any(message in err_msg for message in check_message_contains)
        ):
            # In this case, the error is just logged, suppressed and default
            # value returned
            if logger is not None:
                logger.exception(wrap_error_message_in)

        elif not wrap_error_message_in:
            raise

        else:
            msg = f"{wrap_error_message_in}: {type(err).__name__}: {str(err)}"
            raise type(err)(msg).with_traceback(err.__traceback__)


def _validate_and_fill_defaults(
    retry_on_exception_types: tuple[type[Exception], ...] | None,
    no_retry_on_exception_types: tuple[type[Exception], ...] | None,
    suppress_errors: bool,
    **kwargs: Any,
) -> tuple[tuple[type[Exception], ...], tuple[type[Exception], ...], bool]:
    if retry_on_exception_types is None:
        # If no exception type provided, we catch all errors.
        retry_on_exception_types = (Exception,)
    if not isinstance(retry_on_exception_types, tuple):
        raise ValueError("Expected a tuple of exception types.")

    if no_retry_on_exception_types is not None:
        if not isinstance(no_retry_on_exception_types, tuple):
            raise ValueError("Expected a tuple of non-retriable exception types.")
        if set(no_retry_on_exception_types).intersection(set(retry_on_exception_types)):
            raise ValueError(
                "Same exception type cannot appear in both "
                "`exception_types` and `no_retry_on_exception_types`."
            )

    # `suppress_all_errors` could be a flag to the underlying function
    # when used on instance methods.
    suppress_errors = suppress_errors or kwargs.get("suppress_all_errors", False)
    return retry_on_exception_types, no_retry_on_exception_types or (), suppress_errors


def execute_with_timeout(partial_func: Callable[..., T], timeout: float) -> T:
    """Execute a function in a thread that we can abandon if it takes too long.
    The thread cannot actually be terminated, so the process will keep executing
    after timeout, but not on the main thread.

    Args:
        partial_func: A partial function to execute.  This should either be a
            function that takes no arguments, or a functools.partial function
            with all arguments bound.
        timeout: The timeout in seconds.

    Returns:
        The return value of the partial function when called.
    """
    # since threads cannot return values or raise exceptions in the main thread,
    # we pass it a context dict and have it update it with the return value or
    # exception.
    context_dict = {}

    def execute_partial_with_context(context: dict[str, Any]) -> None:
        """Execute the partial function and update the context dict, which
        will either store the result or exception because threads cannot
        directly return results.

        NOTE: There is a bug specifically when using this with sqlite that the
        database is not initialized and saving within the thread will not
        work if the database connection is initialized outside.
        """
        try:
            context["return_value"] = partial_func()
        except Exception as e:
            context["exception"] = e

    thread = threading.Thread(
        target=partial(execute_partial_with_context, context_dict)
    )
    thread.start()
    thread.join(timeout)
    if thread.is_alive():
        raise TimeoutError("Function timed out")
    if "exception" in context_dict:
        raise context_dict["exception"]
    return context_dict["return_value"]


@contextmanager
def allowed_to_fail(
    allowed_error_types: tuple[Type[Exception], ...] | None = None,
    handler: Callable[[Exception], None] | None = None,
    logger: Logger | None = None,
) -> Iterator[List[Exception | None]]:
    """
    A context manager to catch and handle exceptions of a specific type.
    Args:
        allowed_error_types (Tuple[Type[Exception], ...]): Types of exceptions to catch.
        handler (function, optional): An optional function to handle the exception.
            If not provided, the exception will be ignored. Defaults to None.
        logger (Logger, optional): An optional logger to log the exception.
    """
    allowed_error_types = allowed_error_types or (Exception,)
    maybe_exception: list[Exception | None] = [None]
    try:
        yield maybe_exception
    except allowed_error_types as e:
        maybe_exception[0] = e
        if handler is not None:
            handler(e)
        if logger is not None:
            logger.exception(e)
